"""Filters that can be applied to a Player."""
# SPDX-License-Identifier: MIT
# Reference to filter meanings can be found in:
# https://github.com/natanbc/lavadsp

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Self

    from .typings import (
        ChannelMix as ChannelMixPayload,
        Distortion as DistortionPayload,
        EQBand as EQBandPayload,
        Filters,
        Karaoke as KaraokePayload,
        LowPass as LowPassPayload,
        Rotation as RotationPayload,
        Timescale as TimescalePayload,
        Tremolo as TremoloPayload,
        Vibrato as VibratoPayload,
    )


__all__ = (
    "ChannelMix",
    "Distortion",
    "EQBand",
    "Equalizer",
    "Filter",
    "Karaoke",
    "LowPass",
    "Rotation",
    "Timescale",
    "Tremolo",
    "Vibrato",
)


@dataclass(unsafe_hash=True)
class EQBand:
    """Represents an `equaliser`_ band.

    .. _equaliser: https://en.wikipedia.org/wiki/Equalization_(audio)

    Attributes
    ----------
    band: :class:`int`
        The band to set the gain of. Must be between ``0`` and ``14``.
    gain: :class:`float`
        The gain to set the band to. Must be between ``-0.25`` and ``1.0``.
    """

    band: int
    gain: float

    @property
    def payload(self) -> EQBandPayload:
        """Generate the raw Lavalink payload for this band."""
        return {"band": self.band, "gain": self.gain}


class Equalizer:
    r"""Represents an `equaliser`_.

    .. _equaliser: https://en.wikipedia.org/wiki/Equalization_(audio)

    .. versionchanged:: 2.1

        Added support for :class:`list`\s of
        :class:`tuple`\[:class:`int`, :class:`float`] and :class:`float` to the
        constructor.

    Attributes
    ----------
    bands: :class:`list`\[:class:`EQBand`]
        The bands to set the gains of.
    """

    __slots__ = ("bands",)

    def __init__(
        self, bands: list[EQBand] | list[tuple[int, float]] | list[float]
    ) -> None:
        self.bands = [self._convert_band(band, i) for i, band in enumerate(bands)]

    @staticmethod
    def _convert_band(band: EQBand | tuple[int, float] | float, i: int) -> EQBand:
        if isinstance(band, EQBand):
            return band
        elif isinstance(band, tuple):
            return EQBand(band=band[0], gain=band[1])
        elif isinstance(
            band, (float, int)
        ):  # pyright: ignore[reportUnnecessaryIsInstance]
            return EQBand(band=i, gain=band)
        else:
            msg = f"Expected EQBand, tuple[int, float] or float, got {band!r}"
            raise TypeError(msg)

    def __eq__(self, other: object) -> bool:
        """Check if this equaliser is equal to another."""
        if not isinstance(other, Equalizer):
            return NotImplemented

        return self.bands == other.bands

    def __repr__(self) -> str:
        """Generate a representation of this equaliser."""
        return f"<Equalizer bands={self.bands!r}>"

    def __hash__(self) -> int:
        """Get the hash of this equaliser."""
        return hash(tuple(self.bands))

    @property
    def payload(self) -> list[EQBandPayload]:
        """Generate the raw Lavalink payload for this equaliser."""
        return [band.payload for band in self.bands]

    @classmethod
    def from_payload(cls, payload: list[EQBandPayload]) -> Self:
        """Generate an :class:`Equalizer` from a Lavalink payload."""
        return cls(
            bands=[EQBand(band=band["band"], gain=band["gain"]) for band in payload]
        )


@dataclass(unsafe_hash=True)
class Karaoke:
    r"""Represents a filter which can be used to eliminate part of a band.

    This usually targets vocals, to sound like karaoke music.

    Attributes
    ----------
    level: :data:`~typing.Optional`\[:class:`float`]
        The level of the karaoke effect. Must be between ``0.0`` and ``1.0``.
        Where ``0.0`` is no effect and ``1.0`` is full effect.
        This defaults to ``1.0``.
    mono_level: :data:`~typing.Optional`\[:class:`float`]
        The level of the mono effect. Must be between ``0.0`` and ``1.0``.
        Where ``0.0`` is no effect and ``1.0`` is full effect.
        This defaults to ``1.0``.
    filter_band: :data:`~typing.Optional`\[:class:`float`]
        The frequency of the filter band in Hz.
        This defaults to ``220.0``.
    filter_width: :data:`~typing.Optional`\[:class:`float`]
        The width of the filter band.
        This defaults to ``100.0``.
    """

    level: float | None = None
    mono_level: float | None = None
    filter_band: float | None = None
    filter_width: float | None = None

    @property
    def payload(self) -> KaraokePayload:
        """Generate the raw Lavalink payload for this filter."""
        data: KaraokePayload = {}

        if self.level is not None:
            data["level"] = self.level

        if self.mono_level is not None:
            data["monoLevel"] = self.mono_level

        if self.filter_band is not None:
            data["filterBand"] = self.filter_band

        if self.filter_width is not None:
            data["filterWidth"] = self.filter_width

        return data

    @classmethod
    def from_payload(cls, payload: KaraokePayload) -> Self:
        """Generate a :class:`Karaoke` from a Lavalink payload."""
        return cls(
            level=payload.get("level"),
            mono_level=payload.get("monoLevel"),
            filter_band=payload.get("filterBand"),
            filter_width=payload.get("filterWidth"),
        )


@dataclass(unsafe_hash=True)
class Timescale:
    r"""Represents a filter which is used to change the speed, pitch and rate of audio.

    Attributes
    ----------
    speed: :data:`~typing.Optional`\[:class:`float`]
        The speed of the audio. Must be at least ``0.0``. ``1.0`` is normal speed.
    pitch: :data:`~typing.Optional`\[:class:`float`]
        The pitch of the audio. Must be at least ``0.0``. ``1.0`` is normal pitch.
    rate: :data:`~typing.Optional`\[:class:`float`]
        The rate of the audio. Must be at least ``0.0``. ``1.0`` is normal rate.
    """

    speed: float | None = None
    pitch: float | None = None
    rate: float | None = None

    @property
    def payload(self) -> TimescalePayload:
        """Generate the raw Lavalink payload for this filter."""
        data: TimescalePayload = {}

        if self.speed is not None:
            data["speed"] = self.speed

        if self.pitch is not None:
            data["pitch"] = self.pitch

        if self.rate is not None:
            data["rate"] = self.rate

        return data

    @classmethod
    def from_payload(cls, payload: TimescalePayload) -> Self:
        """Generate a :class:`Timescale` from a Lavalink payload."""
        return cls(
            speed=payload.get("speed"),
            pitch=payload.get("pitch"),
            rate=payload.get("rate"),
        )


@dataclass(unsafe_hash=True)
class Tremolo:
    r"""Represents a filter which can be used to add a `tremolo`_ effect to audio.

    Tremolo oscillates the volume of the audio.

    .. _tremolo: https://en.wikipedia.org/wiki/Tremolo

    Attributes
    ----------
    frequency: :data:`~typing.Optional`\[:class:`float`]
        The frequency of the tremolo effect. Must be at least ``0.0``.
        This defaults to ``2.0``.
    depth: :data:`~typing.Optional`\[:class:`float`]
        The depth of the tremolo effect. Must be between ``0.0`` and ``1.0``.
        Where ``0.0`` is no effect and ``1.0`` is full effect.
        This defaults to ``0.5``.
    """

    frequency: float | None = None
    depth: float | None = None

    @property
    def payload(self) -> TremoloPayload:
        """Generate the raw Lavalink payload for this filter."""
        data: TremoloPayload = {}

        if self.frequency is not None:
            data["frequency"] = self.frequency

        if self.depth is not None:
            data["depth"] = self.depth

        return data

    @classmethod
    def from_payload(cls, payload: TremoloPayload) -> Self:
        """Generate a :class:`Tremolo` from a Lavalink payload."""
        return cls(
            frequency=payload.get("frequency"),
            depth=payload.get("depth"),
        )


@dataclass(repr=True)
class Vibrato:
    r"""Represents a filter which can be used to add a `vibrato`_ effect to audio.

    Vibrato oscillates the pitch of the audio.

    .. _vibrato: https://en.wikipedia.org/wiki/Vibrato

    Attributes
    ----------
    frequency: :data:`~typing.Optional`\[:class:`float`]
        The frequency of the vibrato effect. Must be at least ``0.0``.
        This defaults to ``2.0``.
    depth: :data:`~typing.Optional`\[:class:`float`]
        The depth of the vibrato effect. Must be between ``0.0`` and ``1.0``.
        Where ``0.0`` is no effect and ``1.0`` is full effect.
        This defaults to ``0.5``.
    """

    frequency: float | None = None
    depth: float | None = None

    @property
    def payload(self) -> VibratoPayload:
        """Generate the raw Lavalink payload for this filter."""
        data: VibratoPayload = {}

        if self.frequency is not None:
            data["frequency"] = self.frequency

        if self.depth is not None:
            data["depth"] = self.depth

        return data

    @classmethod
    def from_payload(cls, payload: VibratoPayload) -> Self:
        """Generate a :class:`Vibrato` from a Lavalink payload."""
        return cls(
            frequency=payload.get("frequency"),
            depth=payload.get("depth"),
        )


@dataclass(unsafe_hash=True)
class Rotation:
    r"""Represents a filter which can be used to add a rotating effect to audio.

    An example can be with `"8D audio" (without the reverb)`_.

    .. _"8D audio" (without the reverb): https://youtu.be/QB9EB8mTKcc>

    Attributes
    ----------
    rotation_hz: :data:`~typing.Optional`\[:class:`float`]
        The rotation speed in Hz. Must be at least ``0.0``.
        ``0.2`` is similar to the example above.
    """

    rotation_hz: float | None = None

    @property
    def payload(self) -> RotationPayload:
        """Generate the raw Lavalink payload for this filter."""
        data: RotationPayload = {}

        if self.rotation_hz is not None:
            data["rotationHz"] = self.rotation_hz

        return data

    @classmethod
    def from_payload(cls, payload: RotationPayload) -> Self:
        """Generate a :class:`Rotation` from a Lavalink payload."""
        return cls(
            rotation_hz=payload.get("rotationHz"),
        )


@dataclass(unsafe_hash=True)
class Distortion:
    r"""Represents a filter which can be used to add a distortion effect to audio.

    This applies sine, cosine and tangent distortion to the audio.
    The formula that is used::

        sample_sin = sin_offset + math.sin(sample * sin_scale);
        sample_cos = cos_offset + math.cos(sample * cos_scale);
        sample_tan = tan_offset + math.tan(sample * tan_scale);
        sample = max(
            -1,
            min(
                1,
                offset +
                scale * sample_sin * sample_cos * sample_tan
            )
        )

    Attributes
    ----------
    sin_offset: :data:`~typing.Optional`\[:class:`float`]
        The offset of the sine distortion.
        This defaults to ``0.0``.
    sin_scale: :data:`~typing.Optional`\[:class:`float`]
        The scale of the sine distortion.
        This defaults to ``1.0``.
    cos_offset: :data:`~typing.Optional`\[:class:`float`]
        The offset of the cosine distortion.
        This defaults to ``0.0``.
    cos_scale: :data:`~typing.Optional`\[:class:`float`]
        The scale of the cosine distortion.
        This defaults to ``1.0``.
    tan_offset: :data:`~typing.Optional`\[:class:`float`]
        The offset of the tangent distortion.
        This defaults to ``0.0``.
    tan_scale: :data:`~typing.Optional`\[:class:`float`]
        The scale of the tangent distortion.
        This defaults to ``1.0``.
    offset: :data:`~typing.Optional`\[:class:`float`]
        The offset of the distortion.
        This defaults to ``0.0``.
    scale: :data:`~typing.Optional`\[:class:`float`]
        The scale of the distortion.
        This defaults to ``1.0``.
    """

    sin_offset: float | None = None
    sin_scale: float | None = None
    cos_offset: float | None = None
    cos_scale: float | None = None
    tan_offset: float | None = None
    tan_scale: float | None = None
    offset: float | None = None
    scale: float | None = None

    @property
    def payload(self) -> DistortionPayload:
        """Generate the raw Lavalink payload for this filter."""
        data: DistortionPayload = {}

        if self.sin_offset is not None:
            data["sinOffset"] = self.sin_offset

        if self.sin_scale is not None:
            data["sinScale"] = self.sin_scale

        if self.cos_offset is not None:
            data["cosOffset"] = self.cos_offset

        if self.cos_scale is not None:
            data["cosScale"] = self.cos_scale

        if self.tan_offset is not None:
            data["tanOffset"] = self.tan_offset

        if self.tan_scale is not None:
            data["tanScale"] = self.tan_scale

        if self.offset is not None:
            data["offset"] = self.offset

        if self.scale is not None:
            data["scale"] = self.scale

        return data

    @classmethod
    def from_payload(cls, payload: DistortionPayload) -> Self:
        """Generate a :class:`Distortion` from a Lavalink payload."""
        return cls(
            sin_offset=payload.get("sinOffset"),
            sin_scale=payload.get("sinScale"),
            cos_offset=payload.get("cosOffset"),
            cos_scale=payload.get("cosScale"),
            tan_offset=payload.get("tanOffset"),
            tan_scale=payload.get("tanScale"),
            offset=payload.get("offset"),
            scale=payload.get("scale"),
        )


@dataclass(unsafe_hash=True)
class ChannelMix:
    r"""Represents a filter which can be used to mix the audio channels.

    Setting all of these to ``0.5`` will make the audio mono.

    Attributes
    ----------
    left_to_left: :data:`~typing.Optional`\[:class:`float`]
        The amount of the left channel to mix into the left channel.
        This defaults to ``1.0``.
    left_to_right: :data:`~typing.Optional`\[:class:`float`]
        The amount of the left channel to mix into the right channel.
        This defaults to ``0.0``.
    right_to_left: :data:`~typing.Optional`\[:class:`float`]
        The amount of the right channel to mix into the left channel.
        This defaults to ``0.0``.
    right_to_right: :data:`~typing.Optional`\[:class:`float`]
        The amount of the right channel to mix into the right channel.
        This defaults to ``1.0``.
    """

    left_to_left: float | None = None
    left_to_right: float | None = None
    right_to_left: float | None = None
    right_to_right: float | None = None

    @property
    def payload(self) -> ChannelMixPayload:
        """Generate the raw Lavalink payload for this filter."""
        data: ChannelMixPayload = {}

        if self.left_to_left is not None:
            data["leftToLeft"] = self.left_to_left

        if self.left_to_right is not None:
            data["leftToRight"] = self.left_to_right

        if self.right_to_left is not None:
            data["rightToLeft"] = self.right_to_left

        if self.right_to_right is not None:
            data["rightToRight"] = self.right_to_right

        return data

    @classmethod
    def from_payload(cls, payload: ChannelMixPayload) -> Self:
        """Generate a :class:`ChannelMix` from a Lavalink payload."""
        return cls(
            left_to_left=payload.get("leftToLeft"),
            left_to_right=payload.get("leftToRight"),
            right_to_left=payload.get("rightToLeft"),
            right_to_right=payload.get("rightToRight"),
        )


@dataclass(unsafe_hash=True)
class LowPass:
    r"""Represents a filter which can be used to apply a `low pass filter` to audio.

    High frequencies are suppressed, while low frequencies are passed through.

    .. _low pass filter: http://phrogz.net/js/framerate-independent-low-pass-filter.html

    Attributes
    ----------
    smoothing: :data:`~typing.Optional`\[:class:`float`]
        The smoothing of the low pass filter.
        This defaults to ``0.0``.
    """

    smoothing: float | None = None

    @property
    def payload(self) -> LowPassPayload:
        """Generate the raw Lavalink payload for this filter."""
        data: LowPassPayload = {}

        if self.smoothing is not None:
            data["smoothing"] = self.smoothing

        return data

    @classmethod
    def from_payload(cls, payload: LowPassPayload) -> Self:
        """Generate a :class:`LowPass` from a Lavalink payload."""
        return cls(smoothing=payload.get("smoothing"))


class Filter:
    r"""Represents a filter which can be applied to audio.

    .. container:: operations

        .. describe:: x | y

            Merges two filters together, favouring attributes from y.

        .. describe:: x |= y

            Merges two filters together, favouring attributes from x, assigning to x.

        .. describe:: x & y

            Merges two filters together, favouring attributes from x.

        .. describe:: x &= y

            Merges two filters together, favouring attributes from y, assigning to x.

    .. versionchanged:: 2.1

        ``equalizer`` now accepts a :class:`list` of any of the following:

        - :class:`tuple`\[:class:`int`, :class:`float`\]
        - :class:`float`
        - :class:`EQBand`

    Attributes
    ----------
    equalizer: :data:`~typing.Optional`\[:class:`Equalizer`]
        The equalizer to use.
    karaoke: :data:`~typing.Optional`\[:class:`Karaoke`]
        The karaoke filter to use.
    timescale: :data:`~typing.Optional`\[:class:`Timescale`]
        The timescale filter to use.
    tremolo: :data:`~typing.Optional`\[:class:`Tremolo`]
        The tremolo filter to use.
    vibrato: :data:`~typing.Optional`\[:class:`Vibrato`]
        The vibrato filter to use.
    rotation: :data:`~typing.Optional`\[:class:`Rotation`]
        The rotation filter to use.
    distortion: :data:`~typing.Optional`\[:class:`Distortion`]
        The distortion filter to use.
    channel_mix: :data:`~typing.Optional`\[:class:`ChannelMix`]
        The channel mix filter to use.
    low_pass: :data:`~typing.Optional`\[:class:`LowPass`]
        The low pass filter to use.
    volume: :data:`~typing.Optional`\[:class:`float`]
        The volume to use.
    """

    __slots__ = (
        "equalizer",
        "karaoke",
        "timescale",
        "tremolo",
        "vibrato",
        "rotation",
        "distortion",
        "channel_mix",
        "low_pass",
        "volume",
    )

    def __init__(
        self,
        *,
        equalizer: Equalizer
        | list[tuple[int, float]]
        | list[float]
        | list[EQBand]
        | None = None,
        karaoke: Karaoke | None = None,
        timescale: Timescale | None = None,
        tremolo: Tremolo | None = None,
        vibrato: Vibrato | None = None,
        rotation: Rotation | None = None,
        distortion: Distortion | None = None,
        channel_mix: ChannelMix | None = None,
        low_pass: LowPass | None = None,
        volume: float | None = None,
    ) -> None:
        self.equalizer: Equalizer | None = self._convert_equalizer(equalizer)
        self.karaoke: Karaoke | None = karaoke
        self.timescale: Timescale | None = timescale
        self.tremolo: Tremolo | None = tremolo
        self.vibrato: Vibrato | None = vibrato
        self.rotation: Rotation | None = rotation
        self.distortion: Distortion | None = distortion
        self.channel_mix: ChannelMix | None = channel_mix
        self.low_pass: LowPass | None = low_pass
        self.volume: float | None = volume

    def _convert_equalizer(
        self,
        equalizer: Equalizer
        | list[tuple[int, float]]
        | list[float]
        | list[EQBand]
        | None,
    ) -> Equalizer | None:
        if equalizer is None:
            return None

        if isinstance(equalizer, list):
            return Equalizer(equalizer)

        return equalizer

    def __hash__(self) -> int:
        """Get the hash of this filter."""
        return hash(
            (
                self.equalizer,
                self.karaoke,
                self.timescale,
                self.tremolo,
                self.vibrato,
                self.rotation,
                self.distortion,
                self.channel_mix,
                self.low_pass,
                self.volume,
            )
        )

    def __eq__(self, other: object) -> bool:
        """Check if two filters are equal."""
        if not isinstance(other, Filter):
            return NotImplemented

        return (
            self.equalizer == other.equalizer
            and self.karaoke == other.karaoke
            and self.timescale == other.timescale
            and self.tremolo == other.tremolo
            and self.vibrato == other.vibrato
            and self.rotation == other.rotation
            and self.distortion == other.distortion
            and self.channel_mix == other.channel_mix
            and self.low_pass == other.low_pass
            and self.volume == other.volume
        )

    def __repr__(self) -> str:
        """Generate the string representation of this filter."""
        slots = self.__slots__

        attrs = [
            f"{attr}={getattr(self, attr)!r}"
            for attr in slots
            if getattr(self, attr) is not None
        ]

        return f"<Filter {' '.join(attrs)}>" if attrs else "<Filter>"

    @property
    def payload(self) -> Filters:
        """Generate the raw Lavalink payload for this filter."""
        payload: Filters = {}

        if self.equalizer:
            payload["equalizer"] = self.equalizer.payload

        if self.karaoke:
            payload["karaoke"] = self.karaoke.payload

        if self.timescale:
            payload["timescale"] = self.timescale.payload

        if self.tremolo:
            payload["tremolo"] = self.tremolo.payload

        if self.vibrato:
            payload["vibrato"] = self.vibrato.payload

        if self.rotation:
            payload["rotation"] = self.rotation.payload

        if self.distortion:
            payload["distortion"] = self.distortion.payload

        if self.channel_mix:
            payload["channelMix"] = self.channel_mix.payload

        if self.low_pass:
            payload["lowPass"] = self.low_pass.payload

        if self.volume:
            payload["volume"] = self.volume

        return payload

    @classmethod
    def from_payload(cls, data: Filters) -> Self:
        """Generate a filter from a raw Lavalink payload."""
        self = cls()

        if "equalizer" in data:
            self.equalizer = Equalizer.from_payload(data["equalizer"])

        if "karaoke" in data:
            self.karaoke = Karaoke.from_payload(data["karaoke"])

        if "timescale" in data:
            self.timescale = Timescale.from_payload(data["timescale"])

        if "tremolo" in data:
            self.tremolo = Tremolo.from_payload(data["tremolo"])

        if "vibrato" in data:
            self.vibrato = Vibrato.from_payload(data["vibrato"])

        if "rotation" in data:
            self.rotation = Rotation.from_payload(data["rotation"])

        if "distortion" in data:
            self.distortion = Distortion.from_payload(data["distortion"])

        if "channelMix" in data:
            self.channel_mix = ChannelMix.from_payload(data["channelMix"])

        if "lowPass" in data:
            self.low_pass = LowPass.from_payload(data["lowPass"])

        if "volume" in data:
            self.volume = data["volume"]

        return self

    def __or__(self, other: Filter) -> Filter:
        """Merge two filters together, favouring attributes from other."""
        if not isinstance(
            other, Filter
        ):  # pyright: ignore[reportUnnecessaryIsInstance]
            msg = f"Expected Filter instance, not {type(other)!r}"
            raise TypeError(msg)

        return Filter(
            equalizer=other.equalizer or self.equalizer,
            karaoke=other.karaoke or self.karaoke,
            timescale=other.timescale or self.timescale,
            tremolo=other.tremolo or self.tremolo,
            vibrato=other.vibrato or self.vibrato,
            rotation=other.rotation or self.rotation,
            distortion=other.distortion or self.distortion,
            channel_mix=other.channel_mix or self.channel_mix,
            low_pass=other.low_pass or self.low_pass,
            volume=other.volume or self.volume,
        )

    def __ior__(self, other: Filter) -> None:
        """Merge two filters together, favouring attributes from other, in place."""
        if not isinstance(
            other, Filter
        ):  # pyright: ignore[reportUnnecessaryIsInstance]
            msg = f"Expected Filter instance, not {type(other)!r}"
            raise TypeError(msg)

        self.equalizer = other.equalizer or self.equalizer
        self.karaoke = other.karaoke or self.karaoke
        self.timescale = other.timescale or self.timescale
        self.tremolo = other.tremolo or self.tremolo
        self.vibrato = other.vibrato or self.vibrato
        self.rotation = other.rotation or self.rotation
        self.distortion = other.distortion or self.distortion
        self.channel_mix = other.channel_mix or self.channel_mix
        self.low_pass = other.low_pass or self.low_pass
        self.volume = other.volume or self.volume

    def __and__(self, other: Filter) -> Filter:
        """Merge two filters together, favouring attributes from self."""
        if not isinstance(
            other, Filter
        ):  # pyright: ignore[reportUnnecessaryIsInstance]
            msg = f"Expected Filter instance, not {type(other)!r}"
            raise TypeError(msg)

        return Filter(
            equalizer=self.equalizer or other.equalizer,
            karaoke=self.karaoke or other.karaoke,
            timescale=self.timescale or other.timescale,
            tremolo=self.tremolo or other.tremolo,
            vibrato=self.vibrato or other.vibrato,
            rotation=self.rotation or other.rotation,
            distortion=self.distortion or other.distortion,
            channel_mix=self.channel_mix or other.channel_mix,
            low_pass=self.low_pass or other.low_pass,
            volume=self.volume or other.volume,
        )

    def __iand__(self, other: Filter) -> None:
        """Merge two filters together, favouring attributes from self, in place."""
        if not isinstance(
            other, Filter
        ):  # pyright: ignore[reportUnnecessaryIsInstance]
            msg = f"Expected Filter instance, not {type(other)!r}"
            raise TypeError(msg)

        self.equalizer = self.equalizer or other.equalizer
        self.karaoke = self.karaoke or other.karaoke
        self.timescale = self.timescale or other.timescale
        self.tremolo = self.tremolo or other.tremolo
        self.vibrato = self.vibrato or other.vibrato
        self.rotation = self.rotation or other.rotation
        self.distortion = self.distortion or other.distortion
        self.channel_mix = self.channel_mix or other.channel_mix
        self.low_pass = self.low_pass or other.low_pass
        self.volume = self.volume or other.volume


# TODO: people like easy default filters, add some default EQ and combo filters



"""Type variables used in mafic."""
# SPDX-License-Identifier: MIT
# This was originally made to avoid the import cycle of
# mafic.pool -> mafic.node -> mafic.pool
# pool wants Node, node wants ClientT

from typing import TypeVar

from .__libraries import Client

__all__ = ("ClientT",)
ClientT = TypeVar("ClientT", bound=Client)
r"""A type hint for a client that is (optionally a subclass of) your library's client.

- :class:`dpy:discord.Client` for discord.py.
- :class:`nextcord.Client` for nextcord.
- :class:`disnake.Client` for disnake.
- :class:`pyc:discord.Client` for py-cord.

This is used in :class:`.NodePool` to type hint the client used to create
:class:`Node`\s and :class:`Player`\s. :attr:`Node.client` will be of this type.
"""



# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING, List, TypedDict, Union

if TYPE_CHECKING:
    from .common import Filters, VoiceStateRequest

__all__ = (
    "DecodeTrackParams",
    "OutgoingMessage",
    "OutgoingParams",
    "TrackLoadParams",
    "UnmarkAddressPayload",
    "UpdatePlayerPayload",
    "UpdatePlayerParams",
    "UpdateSessionPayload",
)


class UpdateSessionPayload(TypedDict, total=False):
    # V3
    resumingKey: str | None
    # V4
    resuming: bool
    timeout: int


class UpdatePlayerPayload(TypedDict, total=False):
    encodedTrack: str | None
    identifier: str
    position: int
    endTime: int
    volume: int
    paused: bool
    filters: Filters
    voice: VoiceStateRequest


class UnmarkAddressPayload(TypedDict):
    address: str


class UpdatePlayerParams(TypedDict):
    noReplace: str


class TrackLoadParams(TypedDict):
    identifier: str


class DecodeTrackParams(TypedDict):
    encodedTrack: str


OutgoingMessage = Union[
    UpdatePlayerPayload, UpdateSessionPayload, List[str], UnmarkAddressPayload
]


OutgoingParams = Union[UpdatePlayerParams, TrackLoadParams, DecodeTrackParams]



"""The Lavalink route planner API."""
# SPDX-License-Identifier: MIT

from __future__ import annotations

from abc import ABC
from datetime import datetime, timezone
from enum import Enum
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from .typings import (
        BalancingIPRouteDetails,
        BaseDetails,
        FailingIPAddress,
        IPBlock as IPBlockPayload,
        NanoIPRouteDetails,
        RotatingIPRouteDetails,
        RotatingNanoIPRouteDetails,
    )

__all__ = (
    "IPRoutePlannerType",
    "IPBlockType",
    "IPBlock",
    "FailingAddress",
    "BaseIPRoutePlannerStatus",
    "BalancingIPRoutePlannerStatus",
    "NanoIPRoutePlannerStatus",
    "RotatingIPRoutePlannerStatus",
    "RotatingNanoIPRoutePlannerStatus",
    "RoutePlannerStatus",
)


class IPRoutePlannerType(Enum):
    """The type of IP route planner.

    More info about what planner to use can be found in the lavalink `docs`_.

    .. _docs: https://github.com/freyacodes/Lavalink/blob/master/ROUTEPLANNERS.md
    """

    ROTATING_IP = "RotatingIPRoutePlanner"
    """IP address used is switched on ban.

    This represents ``RotateOnBan`` in config."""

    NANO_IP = "NanoIPRoutePlanner"
    """IP address used is switched on clock update.

    This reprersents ``NanoSwitch`` in config.
    """

    ROTATING_NANO_IP = "RotatingNanoIPRoutePlanner"
    """IP address used is switched on clock update, rotates to a different block on ban.

    This represents ``RotatingNanoSwitch`` in config.
    """

    BALANCING_IP = "BalancingIPRoutePlanner"
    """IP address used is selected randomly.

    This represents ``LoadBalance`` in config.
    """


class IPBlockType(Enum):
    """The type of IP block."""

    V4 = "Inet4Address"
    """Represents an `IPv4`_ address.

    .. _IPv4: https://en.wikipedia.org/wiki/Internet_Protocol_version_4
    """

    V6 = "Inet6Address"
    """Represents an `IPv6`_ address.

    .. _IPv6: https://en.wikipedia.org/wiki/IPv6
    """


class IPBlock:
    """Represents an IP block.

    Attributes
    ----------
    size: :class:`int`
        The size of the block - the number of addresses in the block.
    type: :class:`IPBlockType`
        The type of the block.
    """

    __slots__ = ("size", "type")

    def __init__(self, data: IPBlockPayload) -> None:
        self.size: int = int(data["size"])
        self.type: IPBlockType = IPBlockType(data["type"])


class FailingAddress:
    """Represents a failing IP address.

    Attributes
    ----------
    address: :class:`str`
        The IP address.
    time: :class:`datetime.datetime`
        The time the address was added to the list of failing addresses.
    """

    __slots__ = ("address", "time")

    def __init__(self, data: FailingIPAddress) -> None:
        self.address: str = data["address"]
        self.time: datetime = datetime.fromtimestamp(
            data["failingTimestamp"], tz=timezone.utc
        )


class BaseIPRoutePlannerStatus(ABC):
    r"""An :term:`abstract base class` representing the status of an IP route planner.

    Attributes
    ----------
    failing_addresses: :class:`list`\[:class:`FailingAddress`]
        The list of failing addresses.
    ip_block: :class:`IPBlock`
        The IP block.
    type: :class:`IPRoutePlannerType`
        The type of route planner.
    """

    __slots__ = ("failing_addresses", "ip_block")

    type: IPRoutePlannerType

    def __init__(self, data: BaseDetails) -> None:
        self.ip_block: IPBlock = IPBlock(data["ipBlock"])
        self.failing_addresses: list[FailingAddress] = [
            FailingAddress(addr) for addr in data["failingAddresses"]
        ]


class RotatingIPRoutePlannerStatus(BaseIPRoutePlannerStatus):
    r"""Represents the status of a rotating IP route planner.

    Attributes
    ----------
    type: :data:`~typing.Literal`\[:attr:`IPRoutePlannerType.ROTATING_IP`]
        The type of route planner. This will always be
        :attr:`IPRoutePlannerType.ROTATING_IP`.
    current_address: :class:`str`
        The current IP address.
    ip_index: :class:`int`
        The offset of the current IP address.
    rotate_index: :class:`int`
        The number of rotations.
    """

    __slots__ = ("current_address", "ip_index", "rotate_index")

    type = IPRoutePlannerType.ROTATING_IP

    def __init__(self, data: RotatingIPRouteDetails) -> None:
        super().__init__(data)
        self.rotate_index: int = int(data["rotateIndex"])
        self.ip_index: int = int(data["ipIndex"])
        self.current_address: str = data["currentAddress"]


class NanoIPRoutePlannerStatus(BaseIPRoutePlannerStatus):
    r"""Represents the status of a nano IP route planner.

    Attributes
    ----------
    type: :data:`~typing.Literal`\[:attr:`IPRoutePlannerType.NANO_IP`]
        The type of route planner. This will always be
        :attr:`IPRoutePlannerType.NANO_IP`.
    current_address_index: :class:`int`
        The index of the current address.
    """

    __slots__ = ("current_address_index",)

    type = IPRoutePlannerType.NANO_IP

    def __init__(self, data: NanoIPRouteDetails) -> None:
        super().__init__(data)
        self.current_address_index: int = int(data["currentAddressIndex"])


class RotatingNanoIPRoutePlannerStatus(BaseIPRoutePlannerStatus):
    r"""Represents the status of a rotating nano IP route planner.

    Attributes
    ----------
    type: :data:`~typing.Literal`\[:attr:`IPRoutePlannerType.ROTATING_NANO_IP`]
        The type of route planner. This will always be
        :attr:`IPRoutePlannerType.ROTATING_NANO_IP`.
    block_index: :class:`int`
        The index of the current block.
    current_address_index: :class:`int`
        The index of the current address.
    """

    __slots__ = ("block_index", "current_address_index")

    type = IPRoutePlannerType.ROTATING_NANO_IP

    def __init__(self, data: RotatingNanoIPRouteDetails) -> None:
        super().__init__(data)
        self.block_index: int = int(data["blockIndex"])
        self.current_address_index: int = int(data["currentAddressIndex"])


class BalancingIPRoutePlannerStatus(BaseIPRoutePlannerStatus):
    r"""Represents the status of a balancing IP route planner.

    Attributes
    ----------
    type: :data:`~typing.Literal`\[:attr:`IPRoutePlannerType.BALANCING_IP`]
        The type of route planner. This will always be
        :attr:`IPRoutePlannerType.BALANCING_IP`.
    current_address_index: :class:`int`
        The index of the current address.
    ip_index: :class:`int`
        The offset of the current IP block in the IP block list.
    rotate_index: :class:`int`
        The number of rotations.
    """

    __slots__ = ("current_address_index", "ip_index", "rotate_index")

    type = IPRoutePlannerType.BALANCING_IP

    def __init__(self, data: BalancingIPRouteDetails) -> None:
        super().__init__(data)


RoutePlannerStatus = Union[
    RotatingIPRoutePlannerStatus,
    NanoIPRoutePlannerStatus,
    RotatingNanoIPRoutePlannerStatus,
    BalancingIPRoutePlannerStatus,
]
"""Represents the status of an IP route planner. This can be one of the following:

- :class:`RotatingIPRoutePlannerStatus`
- :class:`NanoIPRoutePlannerStatus`
- :class:`RotatingNanoIPRoutePlannerStatus`
- :class:`BalancingIPRoutePlannerStatus`

To determine the type of route planner, check the :attr:`BaseIPRoutePlannerStatus.type`
attribute.
"""



"""Contains a decorator to merge properties and classmethods."""
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING, Generic, TypeVar

if TYPE_CHECKING:
    from collections.abc import Callable

T = TypeVar("T")
R = TypeVar("R")

__all__ = ("classproperty",)


class _ClassPropertyDescriptor(Generic[T, R]):
    """A descriptor that mimics the behavior of a property, but for classmethods."""

    def __init__(self, fget: classmethod[T, ..., R] | staticmethod[..., R]) -> None:
        self.fget = fget

    def __get__(self, instance: T, owner: type[T] | None = None) -> R:
        if owner is None:
            owner = type(instance)
        return self.fget.__get__(instance, owner)()


def classproperty(
    func: Callable[[T], R] | classmethod[T, ..., R] | staticmethod[..., R]
) -> _ClassPropertyDescriptor[T, R]:
    """Contains a decorator to mimic the behavior of a property, but for classmethods.

    Parameters
    ----------
    func:
        The function to decorate.
    """
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return _ClassPropertyDescriptor(func)



"""Objects for dispatched events via the client."""
# SPDX-License-Identifier: MIT
# pyright: reportImportCycles=false
# Player import.

from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Generic, TypeVar

from .__libraries import VoiceProtocol

if TYPE_CHECKING:
    from .track import Track
    from .typings import (
        LavalinkException,
        TrackEndEvent as TrackEndEventPayload,
        TrackExceptionEvent as TrackExceptionEventPayload,
        TrackStuckEvent as TrackStuckEventPayload,
        WebSocketClosedEvent as WebSocketClosedEventPayload,
    )


# This needs HKTs in python - as Player is generic on ClientT.
PlayerT = TypeVar("PlayerT", bound=VoiceProtocol)

__all__ = (
    "EndReason",
    "TrackEndEvent",
    "TrackExceptionEvent",
    "TrackStartEvent",
    "TrackStuckEvent",
    "WebSocketClosedEvent",
)


class EndReason(str, Enum):
    """Represents the reason why a track ended."""

    FINISHED = "finished"
    """The track finished playing."""

    LOAD_FAILED = "loadFailed"
    """The track failed to load."""

    STOPPED = "stopped"
    """The track was stopped."""

    REPLACED = "replaced"
    """The track was replaced."""

    CLEANUP = "cleanup"
    """The track was cleaned up."""


class WebSocketClosedEvent(Generic[PlayerT]):
    """Represents an event when the connection to Discord is lost.

    Attributes
    ----------
    code: :class:`int`
        The close code.
        Find what this can be in the Discord `docs`_.

        .. _docs: https://discord.dev/topics/opcodes-and-status-codes#close-event-codes.
    reason: :class:`str`
        The close reason.
    by_discord: :class:`bool`
        Whether the close was initiated by Discord.
    player: :class:`Player`
        The player that the event was dispatched from.
    """

    __slots__ = ("code", "reason", "by_discord", "player")

    def __init__(
        self, *, payload: WebSocketClosedEventPayload, player: PlayerT
    ) -> None:
        self.code: int = payload["code"]
        self.reason: str = payload["reason"]
        self.by_discord: bool = payload["byRemote"]
        self.player: PlayerT = player

    def __repr__(self) -> str:
        """Get a string representation of the event."""
        return (
            f"<WebSocketClosedEvent code={self.code} reason={self.reason!r} "
            f"by_discord={self.by_discord}>"
        )


class TrackStartEvent(Generic[PlayerT]):
    """Represents an event when a track starts playing.

    Attributes
    ----------
    track: :class:`Track`
        The track that started playing.
    player: :class:`Player`
        The player that the event was dispatched from.
    """

    __slots__ = ("track", "player")

    def __init__(self, *, track: Track, player: PlayerT) -> None:
        self.track: Track = track
        self.player: PlayerT = player

    def __repr__(self) -> str:
        """Get a string representation of the event."""
        return f"<TrackStartEvent track={self.track!r}>"


class TrackEndEvent(Generic[PlayerT]):
    """Represents an event when a track ends.

    Attributes
    ----------
    track: :class:`Track`
        The track that ended.
    reason: :class:`EndReason`
        The reason why the track ended.
    player: :class:`Player`
        The player that the event was dispatched from.
    """

    __slots__ = ("track", "reason", "player")

    def __init__(
        self, *, track: Track, payload: TrackEndEventPayload, player: PlayerT
    ) -> None:
        self.track: Track = track
        reason = payload["reason"]
        if reason.isupper():
            reason = reason.lower()
            if reason == "load_failed":
                # Only case this happens, not worth a regex sub.
                reason = "loadFailed"
        self.reason: EndReason = EndReason(reason)
        self.player: PlayerT = player

    def __repr__(self) -> str:
        """Get a string representation of the event."""
        return f"<TrackEndEvent track={self.track!r} reason={self.reason!r}>"


class TrackExceptionEvent(Generic[PlayerT]):
    """Represents an event when an exception occurs while playing a track.

    Attributes
    ----------
    track: :class:`Track`
        The track that caused the exception.
    exception: :class:`Exception`
        The exception that was raised.
    player: :class:`Player`
        The player that the event was dispatched from.
    """

    __slots__ = ("track", "exception", "player")

    def __init__(
        self,
        *,
        track: Track,
        payload: TrackExceptionEventPayload,
        player: PlayerT,
    ) -> None:
        self.track: Track = track
        self.exception: LavalinkException = payload["exception"]
        self.player: PlayerT = player

    def __repr__(self) -> str:
        """Get a string representation of the event."""
        return (
            f"<TrackExceptionEvent track={self.track!r} exception={self.exception!r}>"
        )


class TrackStuckEvent(Generic[PlayerT]):
    """Represents an event when a track gets stuck.

    Attributes
    ----------
    track: :class:`Track`
        The track that got stuck.
    threshold_ms: :class:`int`
        The threshold in milliseconds that was exceeded.
    player: :class:`Player`
        The player that the event was dispatched from.
    """

    __slots__ = ("track", "threshold_ms", "player")

    def __init__(
        self, *, track: Track, payload: TrackStuckEventPayload, player: PlayerT
    ) -> None:
        self.track: Track = track
        self.threshold_ms: int = payload["thresholdMs"]
        self.player: PlayerT = player

    def __repr__(self) -> str:
        """Get a string representation of the event."""
        return (
            f"<TrackStuckEvent track={self.track!r} threshold_ms={self.threshold_ms}>"
        )



"""Contains the warnings shown from Mafic."""
# SPDX-License-Identifier: MIT


__all__ = ("UnsupportedVersionWarning", "UnknownVersionWarning")


class UnsupportedVersionWarning(UserWarning):
    """Represents a warning for an unsupported version of Lavalink."""

    message: str = (
        "The version of Lavalink you are using is not supported by Mafic. "
        "It should still work but not all features are supported."
    )


class UnknownVersionWarning(UserWarning):
    """Represents a warning for an unknown version of Lavalink."""

    message: str = (
        "The version of Lavalink you are using is unknown to Mafic. "
        "It should still work but not all features are supported. "
        "Assuming Lavalink v3.7."
    )



# SPDX-License-Identifier: MIT
# ruff: noqa: PGH003

from __future__ import annotations

from os import getenv
from typing import TYPE_CHECKING, Any

from pkg_resources import (
    DistributionNotFound,
    get_distribution,  # pyright: ignore[reportUnknownVariableType]
)

from .errors import MultipleCompatibleLibraries, NoCompatibleLibraries

__all__ = (
    "Client",
    "Connectable",
    "ExponentialBackoff",
    "Guild",
    "GuildChannel",
    "GuildVoiceStatePayload",
    "MISSING",
    "StageChannel",
    "VoiceChannel",
    "VoiceProtocol",
    "VoiceServerUpdatePayload",
    "dumps",
    "loads",
    "version_info",
)

libraries = ("nextcord", "disnake", "py-cord", "discord.py", "discord")
found: list[str] = []


for library in libraries:
    try:
        get_distribution(library)
    except DistributionNotFound:
        pass
    else:
        found.append(library)


if not getenv("MAFIC_IGNORE_LIBRARY_CHECK"):
    if len(found) == 0:
        raise NoCompatibleLibraries
    elif len(found) > 1:
        raise MultipleCompatibleLibraries(found)
elif found[0] == "nextcord":
    from warnings import simplefilter

    # Ignore RuntimeWarning as we import the warning to filter :}
    simplefilter("ignore", RuntimeWarning)
    from nextcord.health_check import DistributionWarning

    simplefilter("always", RuntimeWarning)

    simplefilter("ignore", DistributionWarning)


library = found[0]


if library == "nextcord":
    from nextcord import (
        Client,
        Guild,
        StageChannel,
        VoiceChannel,
        VoiceProtocol,
        version_info,
    )
    from nextcord.abc import Connectable, GuildChannel
    from nextcord.backoff import ExponentialBackoff
    from nextcord.utils import MISSING

    if TYPE_CHECKING:
        from nextcord.types.voice import (
            GuildVoiceState as GuildVoiceStatePayload,
            VoiceServerUpdate as VoiceServerUpdatePayload,
        )
elif library == "disnake":
    from disnake import (
        Client,
        Guild,
        StageChannel,
        VoiceChannel,
        VoiceProtocol,
        version_info,
    )
    from disnake.abc import Connectable, GuildChannel
    from disnake.backoff import ExponentialBackoff
    from disnake.utils import MISSING

    if TYPE_CHECKING:
        if version_info >= (2, 6):
            from disnake.types.gateway import (
                VoiceServerUpdateEvent as VoiceServerUpdatePayload,
            )
        else:
            from disnake.types.voice import (
                # Pyright rules are unnecessarily long.
                VoiceServerUpdate as VoiceServerUpdatePayload,  # pyright: ignore
            )
        from disnake.types.voice import GuildVoiceState as GuildVoiceStatePayload
else:
    from discord import (
        Client,
        Guild,
        StageChannel,
        VoiceChannel,
        VoiceProtocol,
        version_info,
    )
    from discord.abc import Connectable, GuildChannel
    from discord.backoff import ExponentialBackoff
    from discord.utils import MISSING

    if TYPE_CHECKING:
        from discord.types.voice import (
            GuildVoiceState as GuildVoiceStatePayload,  # noqa: TCH004
            VoiceServerUpdate as VoiceServerUpdatePayload,  # noqa: TCH004
        )


try:
    from orjson import dumps as _dumps, loads

    def dumps(obj: Any) -> str:  # noqa: ANN401
        return _dumps(obj).decode()

except ImportError:
    from json import dumps, loads


if version_info.major != 2:
    msg = f"Mafic requires version 2 of {library}."
    raise RuntimeError(msg)



"""A module containing classes to represent node statistics."""
# SPDX-License-Identifier: MIT

from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .typings import CPU, FrameStats as FrameStatsPayload, Memory, Stats


__all__ = (
    "CPUStats",
    "FrameStats",
    "MemoryStats",
    "NodeStats",
)


class CPUStats:
    """Represents CPU stats for a node.

    Attributes
    ----------
    cores: :class:`int`
        The number of cores the node has.
    system_load: :class:`int`
        The load on the whole system lavalink is on..
    lavalink_load: :class:`int`
        The load Lavalink is using.
    """

    __slots__ = ("cores", "system_load", "lavalink_load")

    def __init__(self, payload: CPU) -> None:
        self.cores: int = payload["cores"]
        self.system_load: int = payload["systemLoad"]
        self.lavalink_load: int = payload["lavalinkLoad"]


class MemoryStats:
    """Represents memory stats for a node.

    Attributes
    ----------
    free: :class:`int`
        The amount of free memory.
    used: :class:`int`
        The amount of used memory.
    allocated: :class:`int`
        The amount of allocated memory.
    reservable: :class:`int`
        The amount of reservable memory for the node. Set by ``-Xmx`` for Java.
    """

    __slots__ = ("free", "used", "allocated", "reservable")

    def __init__(self, payload: Memory) -> None:
        self.free: int = payload["free"]
        self.used: int = payload["used"]
        self.allocated: int = payload["allocated"]
        self.reservable: int = payload["reservable"]


class FrameStats:
    """Represents frame stats for a node.

    Attributes
    ----------
    sent: :class:`int`
        The amount of frames sent.
    nulled: :class:`int`
        The amount of frames nulled.
    deficit: :class:`int`
        The amount of frames deficit.
    """

    __slots__ = ("sent", "nulled", "deficit")

    def __init__(self, payload: FrameStatsPayload) -> None:
        self.sent: int = payload["sent"]
        self.nulled: int = payload["nulled"]
        self.deficit: int = payload["deficit"]


class NodeStats:
    r"""Represents stats for a node.

    Attributes
    ----------
    player_count: :class:`int`
        The amount of players connected to the node.
    playing_player_count: :class:`int`
        The amount of players playing on the node.
    uptime: :class:`datetime.timedelta`
        The uptime of the node.
    memory: :class:`MemoryStats`
        The memory stats of the node.
    cpu: :class:`CPUStats`
        The CPU stats of the node.
    frame_stats: :data:`~typing.Optional`\[:class:`FrameStats`]
        The frame stats of the node.
    """

    __slots__ = (
        "cpu",
        "frame_stats",
        "memory",
        "player_count",
        "playing_player_count",
        "uptime",
    )

    def __init__(self, data: Stats) -> None:
        self.player_count: int = data["players"]
        self.playing_player_count: int = data["playingPlayers"]
        self.uptime: timedelta = timedelta(milliseconds=data["uptime"])
        self.memory: MemoryStats = MemoryStats(data["memory"])
        self.cpu: CPUStats = CPUStats(data["cpu"])
        self.frame_stats: FrameStats | None = (
            FrameStats(data["frameStats"])
            if "frameStats" in data and data["frameStats"] is not None
            else None
        )



"""The Lavalink plugin system."""
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .typings import PluginData


__all__ = ("Plugin",)


class Plugin:
    """Represents a plugin.

    Parameters
    ----------
    data:
        The raw data to use to create the plugin.

    Attributes
    ----------
    name: :class:`str`
        The name of the plugin.
    version: :class:`str`
        The version of the plugin.
    """

    __slots__ = ("name", "version")

    def __init__(self, data: PluginData) -> None:
        self.name: str = data["name"]
        self.version: str = data["version"]



"""The module containing :class:`Track`."""
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Self

    from .typings import TrackInfo, TrackWithInfo

__all__ = ("Track",)


class Track:
    r"""Represents a track.

    Parameters
    ----------
    track_id:
        The ID of the track.
    title:
        The title of the track.
    identifier:
        The identifier of the track.
    uri:
        The URI of the track.
    source:
        The source of the track.
    stream:
        Whether the track is a stream.
    seekable:
        Whether the track is seekable.
    position:
        The current position of the track.
    length:
        The length of the track.

    Attributes
    ----------
    id: :class:`str`
        The ID of the track. This is base64 encoded data used by Lavalink.
    title: :class:`str`
        The title of the track.
    author: :class:`str`
        The author of the track.
    identifier: :class:`str`
        The identifier of the track. This is the ID of the track on the source.
    uri: :data:`~typing.Optional`\[:class:`str`]
        The URI of the track.
    source: :class:`str`
        The source of the track.
    stream: :class:`bool`
        Whether the track is a stream.
    seekable: :class:`bool`
        Whether the track is seekable.
    position: :class:`int`
        The current position of the track.
    length: :class:`int`
        The length of the track.
    artwork_url: :data:`~typing.Optional`\[:class:`str`]
        The artwork URL of the track.
        This is always ``None`` if the node does not use Lavalink v4.

        .. versionadded:: 2.2
    isrc: :data:`~typing.Optional`\[:class:`str`]
        The ISRC of the track.
        This is always ``None`` if the node does not use Lavalink v4.

        .. versionadded:: 2.2
    """

    __slots__ = (
        "author",
        "artwork_url",
        "id",
        "identifier",
        "isrc",
        "length",
        "position",
        "seekable",
        "source",
        "stream",
        "title",
        "uri",
    )

    def __init__(
        self,
        *,
        track_id: str,
        identifier: str,
        seekable: bool,
        author: str,
        length: int,
        stream: bool,
        position: int = 0,
        title: str,
        uri: str | None,
        artwork_url: str | None,
        isrc: str | None,
        source: str,
    ) -> None:
        self.id: str = track_id

        self.title: str = title
        self.author: str = author

        self.identifier: str = identifier
        self.uri: str | None = uri
        self.source: str = source

        self.stream: bool = stream

        self.seekable: bool = seekable

        self.position: int = position
        self.length: int = length

        self.artwork_url: str | None = artwork_url
        self.isrc: str | None = isrc

    @classmethod
    def from_data(cls, *, track: str, info: TrackInfo) -> Self:
        """Create a track from the raw data.

        Parameters
        ----------
        track:
            The ID of the track.
        info:
            The track info.

        Returns
        -------
        :class:`Track`
            The track.
        """
        return cls(
            track_id=track,
            title=info["title"],
            author=info["author"],
            identifier=info["identifier"],
            uri=info["uri"],
            source=info["sourceName"],
            stream=info["isStream"],
            seekable=info["isSeekable"],
            position=info["position"],
            length=info["length"],
            artwork_url=info.get("artworkUrl"),
            isrc=info.get("isrc"),
        )

    @classmethod
    def from_data_with_info(cls, data: TrackWithInfo) -> Self:
        """Create a track from the raw data, bundled with the track and info.

        Parameters
        ----------
        data:
            The track and info.

        Returns
        -------
        :class:`Track`
            The track.
        """
        return cls.from_data(track=data["encoded"], info=data["info"])

    def __repr__(self) -> str:
        """Return the string representation of this track."""
        attrs = (
            ("id", self.id),
            ("title", self.title),
            ("author", self.author),
            ("identifier", self.identifier),
            ("uri", self.uri),
            ("source", self.source),
            ("stream", self.stream),
            ("seekable", self.seekable),
            ("position", self.position),
            ("length", self.length),
        )

        resolved = " ".join(f"{name}={value!r}" for name, value in attrs)
        return f"<{type(self).__name__} {resolved}>"



# SPDX-License-Identifier: MIT
"""A properly typehinted lavalink client for discord.py, nextcord, disnake and py-cord.

:copyright: (c) 2022-present ooliver1
:license: MIT, see LICENSE for more details.
"""

import logging

from . import __libraries
from .errors import *
from .events import *
from .filter import *
from .ip import *
from .node import *
from .player import *
from .playlist import *
from .pool import *
from .region import *
from .search_type import *
from .stats import *
from .strategy import *
from .track import *
from .warnings import *

__title__ = "mafic"
__author__ = "ooliver1"
__license__ = "MIT"
__copyright__ = "Copyright 2022-present ooliver1"
__version__ = "2.10.0"
"""The current version of mafic, using `PEP 440`_ format.

.. _PEP 440: https://peps.python.org/pep-0440/
"""

del __libraries

logging.getLogger(__name__).addHandler(logging.NullHandler())



# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, TypedDict

from .misc import PayloadWithGuild

if TYPE_CHECKING:
    from typing_extensions import NotRequired

__all__ = (
    "Filters",
    "ChannelMix",
    "EQBand",
    "Distortion",
    "Filters",
    "Karaoke",
    "LowPass",
    "Player",
    "PlaylistInfo",
    "Rotation",
    "Timescale",
    "TrackInfo",
    "TrackWithInfo",
    "Tremolo",
    "Vibrato",
    "VoiceState",
    "VoiceStateRequest",
    "Memory",
    "CPU",
    "FrameStats",
    "Stats",
)


class EQBand(TypedDict):
    band: int
    gain: float


class Karaoke(TypedDict, total=False):
    level: float
    monoLevel: float
    filterBand: float
    filterWidth: float


class Timescale(TypedDict, total=False):
    speed: float
    pitch: float
    rate: float


class Tremolo(TypedDict, total=False):
    frequency: float
    depth: float


class Vibrato(TypedDict, total=False):
    frequency: float
    depth: float


class Rotation(TypedDict, total=False):
    rotationHz: float


class Distortion(TypedDict, total=False):
    sinOffset: float
    sinScale: float
    cosOffset: float
    cosScale: float
    tanOffset: float
    tanScale: float
    offset: float
    scale: float


class ChannelMix(TypedDict, total=False):
    leftToLeft: float
    leftToRight: float
    rightToLeft: float
    rightToRight: float


class LowPass(TypedDict, total=False):
    smoothing: float


class Filters(TypedDict, total=False):
    volume: float
    equalizer: list[EQBand]
    karaoke: Karaoke
    timescale: Timescale
    tremolo: Tremolo
    vibrato: Vibrato
    rotation: Rotation
    distortion: Distortion
    channelMix: ChannelMix
    lowPass: LowPass


class Player(PayloadWithGuild):
    track: TrackWithInfo | None
    volume: int
    paused: bool
    voice: VoiceState
    filters: Filters


class VoiceStateRequest(TypedDict):
    token: str
    endpoint: str
    sessionId: str


class VoiceState(VoiceStateRequest):
    connected: bool
    ping: int


class PlaylistInfo(TypedDict):
    name: str
    selectedTrack: int


class TrackInfo(TypedDict):
    identifier: str
    isSeekable: bool
    author: str
    length: int
    isStream: bool
    position: int
    title: str
    uri: str | None
    # TODO: Remove NotRequired when V3 support removed
    artworkUrl: NotRequired[str | None]
    isrc: NotRequired[str | None]
    sourceName: str


class TrackWithInfo(TypedDict):
    encoded: str
    info: TrackInfo


class Memory(TypedDict):
    free: int
    used: int
    allocated: int
    reservable: int


class CPU(TypedDict):
    cores: int
    systemLoad: int
    lavalinkLoad: int


class FrameStats(TypedDict):
    sent: int
    nulled: int
    deficit: int


class Stats(TypedDict):
    op: Literal["stats"]
    players: int
    playingPlayers: int
    uptime: int
    memory: Memory
    cpu: CPU
    # V3 is NotRequired, V4 is None
    frameStats: NotRequired[FrameStats | None]



"""Represents a search type for Lavalink."""
# SPDX-License-Identifier: MIT

from __future__ import annotations

from enum import Enum

__all__ = ("SearchType",)


class SearchType(Enum):
    """Represents a search type for Lavalink.

    .. danger::

        Many of these require plugins to be used.

        `LavaSrc <https://github.com/TopiSenpai/LavaSrc>`_:
            - :attr:`SPOTIFY_SEARCH`
            - :attr:`SPOTIFY_RECOMMENDATIONS`
            - :attr:`APPLE_MUSIC`
            - :attr:`DEEZER_SEARCH`
            - :attr:`DEEZER_ISRC`
            - :attr:`YANDEX`

        `DuncteBot <https://github.com/DuncteBot/skybot-lavalink-plugin>`_:
            - :attr:`SPEECH`
    """

    YOUTUBE = "ytsearch"
    YOUTUBE_MUSIC = "ytmsearch"
    SOUNDCLOUD = "scsearch"

    # Plugins

    # LavaSrc
    SPOTIFY_SEARCH = "spsearch"
    SPOTIFY_RECOMMENDATIONS = "sprec"
    APPLE_MUSIC = "amsearch"
    DEEZER_SEARCH = "dzsearch"
    DEEZER_ISRC = "dzisrc"
    YANDEX_MUSIC = "ymsearch"

    # DuncteBot
    TTS = "speak"



"""A Player is used to connect to a channel."""
# SPDX-License-Identifier: MIT

from __future__ import annotations

import asyncio
import warnings
from asyncio import Event
from collections import OrderedDict
from functools import reduce
from logging import getLogger
from operator import or_
from time import time
from typing import TYPE_CHECKING, Generic, cast

from .__libraries import (
    MISSING,
    GuildChannel,
    StageChannel,
    VoiceChannel,
    VoiceProtocol,
)
from .errors import NoNodesAvailable, PlayerNotConnected
from .events import *
from .filter import Filter
from .pool import NodePool
from .search_type import SearchType
from .track import Track
from .type_variables import ClientT
from .typings import TrackWithInfo

if TYPE_CHECKING:
    from .__libraries import (
        Connectable,
        Guild,
        GuildVoiceStatePayload,
        VoiceServerUpdatePayload,
    )
    from .node import Node
    from .playlist import Playlist
    from .typings import EventPayload, Player as PlayerPayload, PlayerUpdateState


_log = getLogger(__name__)
__all__ = ("Player",)


class Player(VoiceProtocol, Generic[ClientT]):
    """Represents a player for a guild.

    .. note::

        This class is not meant to be instantiated by the user.
        Use ``Connectable.connect()`` from your Discord library to connect.

    Parameters
    ----------
    client:
        The client that the player is associated with.
    channel:
        The voice channel to connect to.
    node:
        The node to use for the player. If not provided, the best node will be used.

    Attributes
    ----------
    client: :data:`~mafic.type_variables.ClientT`
        The client that the player is associated with.
    channel:
        The voice channel that the player is connected to.
        This is an ``abc.Connectable`` from your Discord library.
    guild:
        The guild that the player is associated with.
        This is a ``Guild`` from your Discord library.
    """

    def __init__(
        self,
        client: ClientT,
        channel: Connectable,
        *,
        node: Node[ClientT] | None = None,
    ) -> None:
        self.client: ClientT = client
        self.channel: Connectable = channel

        if not isinstance(self.channel, GuildChannel):
            msg = "Voice channel must be a GuildChannel."
            raise TypeError(msg)

        self.guild: Guild = self.channel.guild
        self.endpoint: str | None = None

        self._node = node

        self._guild_id: int = self.guild.id
        self._session_id: str | None = None
        self._server_state: VoiceServerUpdatePayload | None = None
        self._connected: bool = False
        self._position: int = 0
        self._last_update: int = 0
        self._ping = -1
        self._current: Track | None = None
        self._filters: OrderedDict[str, Filter] = OrderedDict()
        # Used to get the last track for TrackEndEvent.
        self._last_track: Track | None = None
        self._paused: bool = False

        self._voice_server_update_event: Event = Event()
        self._voice_state_update_event: Event = Event()
        self._node_player_ready_event: Event = Event()

    def set_state(self, state: PlayerPayload) -> None:
        """Set the state of the player.

        .. note::

            This is used internally to set the state of the player.
            You should not need to use this.
        """
        self._session_id = state["voice"]["sessionId"]
        self._ping = state["voice"].get("ping", -1)
        self._current = (
            Track.from_data_with_info(state["track"]) if state["track"] else None
        )
        self._filters = OrderedDict({"RESUME": Filter.from_payload(state["filters"])})
        self._paused = state["paused"]
        self._server_state = {
            "token": state["voice"]["token"],
            "endpoint": state["voice"]["endpoint"],
            "guild_id": self._guild_id,
        }
        if state["track"]:
            self._position = state["track"]["info"]["position"]

    def __repr__(self) -> str:
        """Return a string representation of the player."""
        attrs = (
            ("guild_id", self._guild_id),
            ("session_id", self._session_id),
            ("connected", self._connected),
            ("position", self._position),
            ("current", self._current),
        )
        resolved = " ".join(f"{attr}={value!r}" for attr, value in attrs)
        return f"<{type(self).__name__} {resolved}>"

    @property
    def connected(self) -> bool:
        """Whether the player is connected to a voice channel."""
        return self._connected

    @property
    def position(self) -> int:
        """The current position of the player in milliseconds."""
        pos = self._position

        if self._connected and self._current is not None:
            # Add the time since the last update to the position.
            # If the track total time is less than that, use that.
            pos = min(
                self._current.length, pos + int((time() * 1000) - self._last_update)
            )

        return pos

    @property
    def ping(self) -> int:
        """The current ping of the player in milliseconds."""
        return self._ping

    @property
    def node(self) -> Node[ClientT]:
        """The node that the player is connected to."""
        if self._node is None:
            _log.warning(
                "Unable to use best node, player not connected, finding random node.",
                extra={"guild": self._guild_id},
            )
            return NodePool[ClientT].get_random_node()

        return self._node

    @property
    def current(self) -> Track | None:
        """The current track that is playing."""
        return self._current

    @property
    def paused(self) -> bool:
        """Whether the player is paused."""
        return self._paused

    def update_state(self, state: PlayerUpdateState) -> None:
        """Update the player state.

        This is called by the library and usually should not be called by the user.

        Parameters
        ----------
        state:
            The state to update the player with.
        """
        if not self._node_player_ready_event.is_set():
            self._node_player_ready_event.set()

        self._last_update = state["time"]
        self._position = state.get("position", 0)
        self._connected = state["connected"]
        self._ping = state.get("ping", -1)

    # If people are so in love with the VoiceClient interface
    def is_connected(self) -> bool:
        """Whether the player is connected to a voice channel.

        This is an alias for :attr:`connected`.
        """
        return self._connected

    async def _dispatch_player_update(self) -> None:
        """Dispatch a player update to the node."""
        if self._node is None:
            _log.debug("Recieved voice update before node was found.")
            return

        if self._session_id is None:
            _log.debug("Receieved player update before session ID was set.")
            return

        if self._server_state is None:
            _log.debug("Receieved player update before server state was found")
            return

        await self._node.voice_update(
            guild_id=self._guild_id,
            session_id=self._session_id,
            data=self._server_state,
        )

    def dispatch_event(self, data: EventPayload) -> None:
        """Dispatch an event to the player.

        This is called by the library and usually should not be called by the user.

        Parameters
        ----------
        data:
            The event payload to dispatch.
        """
        if data["type"] == "WebSocketClosedEvent":
            event = WebSocketClosedEvent(payload=data, player=self)
            _log.debug("Received websocket closed event: %s", event)
            self.client.dispatch("websocket_closed", event)
        elif data["type"] == "TrackStartEvent":
            track = (
                self._current
                if self.node.version == 3
                else Track.from_data_with_info(cast(TrackWithInfo, data.get("track")))
            )

            if track is None:
                _log.error(
                    "Received track start event but no track was playing, discarding."
                )
                return

            event = TrackStartEvent(player=self, track=track)
            self.client.dispatch("track_start", event)
            _log.debug("Received track start event: %s", event)
            self._last_track = track
        elif data["type"] == "TrackEndEvent":
            track = (
                self._last_track
                if self.node.version == 3
                else Track.from_data_with_info(cast(TrackWithInfo, data.get("track")))
            )

            if track is None:
                _log.error(
                    "Received track end event but no track was playing, discarding."
                )
                return

            event = TrackEndEvent(player=self, track=track, payload=data)
            self.client.dispatch("track_end", event)
            _log.debug("Received track end event: %s", event)

            if data["reason"] != "REPLACED":
                self._current = None
        elif data["type"] == "TrackExceptionEvent":
            track = (
                self._current
                if self.node.version == 3
                else Track.from_data_with_info(cast(TrackWithInfo, data.get("track")))
            )

            if track is None:
                _log.error(
                    "Received track exception event but no track was playing, "
                    "discarding."
                )
                return

            event = TrackExceptionEvent(player=self, track=track, payload=data)
            self.client.dispatch("track_exception", event)
            _log.debug("Received track exception event: %s", event)
        elif data["type"] == "TrackStuckEvent":
            track = (
                self._current
                if self.node.version == 3
                else Track.from_data_with_info(cast(TrackWithInfo, data.get("track")))
            )

            if track is None:
                _log.error(
                    "Received track stuck event but no track was playing, discarding."
                )
                return

            event = TrackStuckEvent(player=self, track=track, payload=data)
            self.client.dispatch("track_stuck", event)
            _log.debug("Received track stuck event: %s", event)
        else:
            # Pyright expects this to never happen, so do I, I really hope.
            # Nobody expects the Spanish Inquisition, neither does pyright.

            event_type = cast(str, data["type"])
            _log.warning("Unknown incoming event type %s", event_type)

    async def on_voice_state_update(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, data: GuildVoiceStatePayload
    ) -> None:
        """Dispatch a voice state update.

        This is called by the library and usually should not be called by the user.

        Parameters
        ----------
        data:
            The voice state update payload.
        """
        before_session_id = self._session_id
        self._session_id = data["session_id"]

        channel_id = data["channel_id"]

        if channel_id is None:  # pyright: ignore[reportUnnecessaryComparison]
            # This can happen and is on disconnect.
            # Not sure why this is typed as always Snowflake.
            await self.disconnect(force=True)
            return

        channel = self.guild.get_channel(int(channel_id))
        if not isinstance(channel, (VoiceChannel, StageChannel)):
            msg = "Channel was not a connectable channel that was recognised."
            raise TypeError(msg)

        self.channel = channel

        if self._session_id != before_session_id:
            await self._dispatch_player_update()

        self._voice_state_update_event.set()

    async def on_voice_server_update(self, data: VoiceServerUpdatePayload) -> None:
        """Handle a voice server update.

        This is called by the library and usually should not be called by the user.

        Parameters
        ----------
        data:
            The voice server update payload.
        """
        # Fetch the best node as we either don't know the best one yet.
        # Or the node we were using was not the best one (endpoint optimisation).
        if (
            self._node is None
            or self._server_state is None
            or self._server_state["endpoint"] != data["endpoint"]
        ):
            _log.debug("Getting best node for player", extra={"guild": self._guild_id})
            self._node = NodePool[ClientT].get_node(
                guild_id=data["guild_id"], endpoint=data["endpoint"]
            )
            _log.debug(
                "Got best node for player: %s",
                self._node.label,
                extra={"guild": self._guild_id},
            )

        self._node.add_player(self._guild_id, self)

        self._guild_id = int(data["guild_id"])
        self._endpoint = data["endpoint"]
        self._server_state = data

        await self._dispatch_player_update()

        self._voice_server_update_event.set()

    async def connect(
        self,
        *,
        timeout: float,
        reconnect: bool,  # noqa: ARG002
        self_mute: bool = False,
        self_deaf: bool = False,
    ) -> None:
        """Connect to a voice channel.

        This is called by your Discord library when using ``Connectable.connect``.

        Parameters
        ----------
        timeout:
            The timeout for the connection.
        reconnect:
            Whether to reconnect to the voice channel if the connection is lost.
        self_mute:
            Whether to mute the bot on connect.
        self_deaf:
            Whether to deafen the bot on connect.
        """
        if not isinstance(self.channel, (VoiceChannel, StageChannel)):
            msg = "Voice channel must be a VoiceChannel or StageChannel."
            raise TypeError(msg)

        # I really don't like how pyright rules are so incredibly long,
        # This has 2 type issues.
        if not NodePool.nodes:  # pyright: ignore   # noqa: PGH003
            self.cleanup()
            raise NoNodesAvailable

        _log.debug("Connecting to voice channel %s", self.channel.id)

        await self.channel.guild.change_voice_state(
            channel=self.channel, self_mute=self_mute, self_deaf=self_deaf
        )
        futures = [
            self._voice_state_update_event.wait(),
            self._voice_server_update_event.wait(),
            self._node_player_ready_event.wait(),
        ]

        ensured = [asyncio.ensure_future(fut) for fut in futures]
        _, pending = await asyncio.wait(
            ensured, timeout=timeout, return_when=asyncio.ALL_COMPLETED
        )

        if len(pending) != 0:
            await self.disconnect(force=True)
            raise asyncio.TimeoutError

    async def disconnect(self, *, force: bool = False) -> None:
        """Disconnect from the voice channel.

        Parameters
        ----------
        force:
            Whether to force the disconnect even if not connected.
        """
        if not self._connected and not force:
            return

        # Handled by Discord
        if self.client.is_closed():
            return

        _log.debug(
            "Disconnecting player and destroying client.",
            extra={"guild": self._guild_id},
        )
        if self._node is not None:
            self._node.remove_player(self.guild.id)
            await self._node.destroy(guild_id=self.guild.id)

        try:
            _log.debug(
                "Disconnecting from voice channel.",
                extra={"guild": self._guild_id},
            )
            await self.guild.change_voice_state(channel=None)
        finally:
            self.cleanup()
            self._connected = False

    def cleanup(self) -> None:
        """Clean up the player.

        This shouldn't be called directly. Instead, use :meth:`disconnect`.
        """
        self._current = None
        self._position = 0
        self._paused = False
        self._ping = 0
        self._channel = None
        self._connected = False

        return super().cleanup()

    async def transfer_to(self, node: Node[ClientT]) -> None:
        """Transfer the player to a new node.

        Parameters
        ----------
        node:
            The node to transfer to.
        """
        if self._node is None:
            raise PlayerNotConnected

        if self._node == node:
            return

        state = await self._node.fetch_player(self.guild.id)

        # Remove from the current node, but no need to destroy.
        self._node.remove_player(self.guild.id)

        old_node = self._node
        self._node = node
        self._node.add_player(self.guild.id, self)

        if self._session_id is None or self._server_state is None:
            msg = "Cannot transfer player with session data."
            raise RuntimeError(msg)

        # We need to update the voice server as the endpoint may have changed.
        await self._node.voice_update(
            guild_id=self._guild_id,
            session_id=self._session_id,
            data=self._server_state,
        )

        # Needed so .update does not fail.
        self._connected = True
        # Update player with all other state.
        # Position, filters, track, etc.
        await self.update(
            track=self._current,
            position=self.position,
            volume=state["volume"],
            pause=self._paused,
            filter=reduce(or_, self._filters.values()) if self._filters else Filter(),
        )

        await old_node.destroy(guild_id=self.guild.id)

    async def destroy(self) -> None:  # noqa: D102
        warnings.warn(
            "Player.destroy is deprecated and will be removed in 3.0.0, "
            "use .disconnect instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        await self.disconnect()

    async def fetch_tracks(
        self, query: str, search_type: SearchType | str = SearchType.YOUTUBE
    ) -> list[Track] | Playlist | None:
        r"""Fetch tracks from the node.

        Parameters
        ----------
        query:
            The query to search for.
        search_type:
            The search type to use.

        Returns
        -------
        :class:`list`\[:class:`Track`]
            A list of tracks if the load type is ``TRACK_LOADED`` or ``SEARCH_RESULT``.
        :class:`Playlist`
            A playlist if the load type is ``PLAYLIST_LOADED``.
        None
            If the load type is ``NO_MATCHES``.

        Notes
        -----
        If a node was not selected due to not being connected, this will use a
        random node.
        """
        node = self.node

        raw_type: str
        if isinstance(search_type, SearchType):
            raw_type = search_type.value
        else:
            raw_type = search_type

        return await node.fetch_tracks(query, search_type=raw_type)

    async def update(
        self,
        *,
        track: Track | str | None = MISSING,
        position: int | None = None,
        end_time: int | None = None,
        volume: int | None = None,
        pause: bool | None = None,
        filter: Filter | None = None,
        replace: bool = False,
    ) -> None:
        """Update the player.

        Parameters
        ----------
        track:
            The track to play. This can be a :class:`Track` or an identifier.

            .. versionchanged:: 2.4
                The track can now be a :class:`str` to play an identifier.
        position:
            The position to start the track at.
        end_time:
            The time to end the track at.
        volume:
            The volume to play the track at.
        pause:
            Whether to pause the track.
        filter:
            The filter to apply to the track.
        replace:
            Whether to replace the current track if one is playing.

        Raises
        ------
        PlayerNotConnected
            If the player is not connected to a voice channel.
        """
        if self._node is None or not self._connected:
            raise PlayerNotConnected

        # v4+ receives the full payload in events, so there is no need to
        # pre-emptyively update the player.
        if track is not None and self.node.version == 3:
            if isinstance(track, str):
                message = (
                    "Lavalink version 3 does not support identifiers "
                    "due to unexpected behaviour with events."
                )
                raise TypeError(message)

            self._current = track

        data = await self._node.update(
            guild_id=self._guild_id,
            track=track,
            position=position,
            end_time=end_time,
            volume=volume,
            pause=pause,
            filter=filter,
            no_replace=not replace,
        )

        if data["track"]:
            self._current = Track.from_data_with_info(data["track"])

        self._volume = data["volume"]
        self._paused = data["paused"]

    async def play(
        self,
        track: Track | str,
        /,
        *,
        start_time: int | None = None,
        end_time: int | None = None,
        volume: int | None = None,
        replace: bool = True,
        pause: bool | None = None,
    ) -> None:
        """Play the given track.

        Parameters
        ----------
        track:
            The track to play. This can be a :class:`Track` or an identifier.

            .. versionchanged:: 2.4
                The track can now be a :class:`str` to play an identifier.
        start_time:
            The position to start the track at.
        end_time:
            The time to end the track at.
        volume:
            The volume to play the track at.
        replace:
            Whether to replace the current track if one is playing.
        pause:
            Whether to pause the track.

        Raises
        ------
        PlayerNotConnected
            If the player is not connected to a voice channel.

        Notes
        -----
        This is a convenience method for :meth:`update`.
        """
        return await self.update(
            track=track,
            position=start_time,
            end_time=end_time,
            volume=volume,
            replace=replace,
            pause=pause,
        )

    async def pause(self, pause: bool = True) -> None:  # noqa: FBT001, FBT002
        """Pause the current track.

        Parameters
        ----------
        pause:
            Whether to pause the track.

        Raises
        ------
        PlayerNotConnected
            If the player is not connected to a voice channel.

        Notes
        -----
        This is a convenience method for :meth:`update`.
        """
        return await self.update(pause=pause)

    async def resume(self) -> None:
        """Resume the current track.

        Raises
        ------
        PlayerNotConnected
            If the player is not connected to a voice channel.

        Notes
        -----
        This is a convenience method for :meth:`pause`, with ``pause`` set to
        ``False``.
        """
        return await self.pause(False)  # noqa: FBT003

    async def stop(self) -> None:
        """Stop the current track.

        Raises
        ------
        PlayerNotConnected
            If the player is not connected to a voice channel.

        Notes
        -----
        This is a convenience method for :meth:`update`.
        """
        await self.update(track=None, replace=True)

    async def _update_filters(self, *, fast_apply: bool) -> None:
        """Update the filters on the player.

        Parameters
        ----------
        fast_apply:
            Whether to seek to the current position after updating the filters.
        """
        await self.update(
            filter=reduce(or_, self._filters.values()) if self._filters else Filter()
        )

        if fast_apply:
            await self.seek(self.position)

    async def add_filter(
        self, filter: Filter, /, *, label: str, fast_apply: bool = False
    ) -> None:
        """Add a filter to the player.

        Parameters
        ----------
        filter:
            The filter to add.
        label:
            The label to use for the filter. This is used to remove the filter
            later.
        fast_apply:
            Whether to seek to the current position after updating the filters.
            This clears Lavalink's internal buffer, which may cause a small
            delay in the audio.

        Raises
        ------
        PlayerNotConnected
            If the player is not connected to a voice channel.
        """
        self._filters[label] = filter

        await self._update_filters(fast_apply=fast_apply)

    async def has_filter(self, label: str) -> bool:
        """Check if the player has a filter with the given label.

        .. versionadded:: 2.1

        Parameters
        ----------
        label:
            The label to check for.

        Returns
        -------
        :class:`bool`
            Whether the player has a filter with the given label.
        """
        return label in self._filters

    async def remove_filter(self, label: str, *, fast_apply: bool = False) -> None:
        """Remove a filter from the player.

        Parameters
        ----------
        label:
            The label of the filter to remove.
        fast_apply:
            Whether to seek to the current position after updating the filters.
            This clears Lavalink's internal buffer, which may cause a small
            delay in the audio.

        Raises
        ------
        PlayerNotConnected
            If the player is not connected to a voice channel.
        ValueError
            If the filter does not exist.
        """
        self._filters.pop(label)

        await self._update_filters(fast_apply=fast_apply)

    async def clear_filters(self, *, fast_apply: bool = False) -> None:
        """Remove all filters from the player.

        Parameters
        ----------
        fast_apply:
            Whether to seek to the current position after updating the filters.
            This clears Lavalink's internal buffer, which may cause a small
            delay in the audio.

        Raises
        ------
        PlayerNotConnected
            If the player is not connected to a voice channel.
        """
        self._filters.clear()

        await self._update_filters(fast_apply=fast_apply)

    async def set_volume(self, volume: int, /) -> None:
        """Set the volume of the player.

        Parameters
        ----------
        volume:
            The volume to set.

        Raises
        ------
        PlayerNotConnected
            If the player is not connected to a voice channel.

        Notes
        -----
        This is a convenience method for :meth:`update`.
        """
        await self.update(volume=volume)

    async def seek(self, position: int, /) -> None:
        """Seek to a position in the current track.

        Parameters
        ----------
        position:
            The position to seek to, in milliseconds.

        Raises
        ------
        PlayerNotConnected
            If the player is not connected to a voice channel.

        Notes
        -----
        This is a convenience method for :meth:`update`.
        """
        return await self.update(position=position)



"""Errors raised by Mafic."""
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Self

    from .typings import ExceptionSeverity, LavalinkException

__all__ = (
    "HTTPBadRequest",
    "HTTPException",
    "HTTPNotFound",
    "HTTPUnauthorized",
    "LibraryCompatibilityError",
    "MaficException",
    "MultipleCompatibleLibraries",
    "NoCompatibleLibraries",
    "NoNodesAvailable",
    "NodeAlreadyConnected",
    "PlayerException",
    "PlayerNotConnected",
    "TrackLoadException",
)


class MaficException(Exception):
    """The base exception class for custom exceptions raised by Mafic."""


class LibraryCompatibilityError(MaficException):
    """An issue occured when trying to find a compatible library."""


class NoCompatibleLibraries(LibraryCompatibilityError):
    """No compatible library was found."""

    def __init__(self) -> None:
        super().__init__(
            "No compatible libraries were found. Please install one of the following: "
            "nextcord, disnake, py-cord, discord.py or discord."
        )


class MultipleCompatibleLibraries(LibraryCompatibilityError):
    """Multiple compatible libraries were found.

    Mafic makes no attempt to assume which library you are using.
    """

    def __init__(self, libraries: list[str]) -> None:
        super().__init__(
            f"Multiple compatible libraries were found: {', '.join(libraries)}. "
            "Please remove all but one of the libraries."
        )


class PlayerException(MaficException):
    """An issue occured when trying to play a track."""


class TrackLoadException(PlayerException):
    r"""An error raised when a track could not be loaded.

    Attributes
    ----------
    message: :class:`str`
        The message returned by the node.
    severity: :data:`~typing.Literal`\[``"COMMON"``, ``"SUSPICIOUS"``, ``"FATAL"``]
        The severity of the error.

        This is lowercase in Lavalink v4.
    cause: :class:`str`
        The cause of the error.
    """

    def __init__(
        self, *, message: str, severity: ExceptionSeverity, cause: str
    ) -> None:
        super().__init__(f"The track could not be loaded: {message} ({severity} error)")

        self.message: str = message
        self.severity: ExceptionSeverity = severity
        self.cause: str = cause

    @classmethod
    def from_data(cls, data: LavalinkException) -> Self:
        """Construct a new TrackLoadException from raw Lavalink data.

        Parameters
        ----------
        data:
            The raw data from Lavalink.

        Returns
        -------
        TrackLoadException
            The constructed exception.
        """
        return cls(
            message=data["message"], severity=data["severity"], cause=data["cause"]
        )


class PlayerNotConnected(PlayerException):
    """An error raised when a player is not connected to a voice channel."""

    def __init__(self) -> None:
        super().__init__("The player is not connected to a voice channel.")


class NodeAlreadyConnected(MaficException):
    """An error raised when a node is already connected to Mafic."""

    def __init__(self) -> None:
        super().__init__("The node is already connected to Mafic.")


class NoNodesAvailable(MaficException):
    """An error raised when no nodes are available to handle a player."""

    def __init__(self) -> None:
        super().__init__("No nodes are available to handle this player.")


class HTTPException(MaficException):
    """An issue occured when trying to make a request to the Lavalink REST API."""

    def __init__(self, status: int, message: str) -> None:
        super().__init__(f"An HTTP error occured: {message} ({status})")

        self.status: int = status
        self.message: str = message


class HTTPBadRequest(HTTPException):
    """An error raised when a bad request is made to the Lavalink REST API."""

    def __init__(self, message: str) -> None:
        super().__init__(400, message)


class HTTPUnauthorized(HTTPException):
    """An error raised when an unauthorized request is made to the Lavalink REST API."""

    def __init__(self, message: str) -> None:
        super().__init__(401, message)


class HTTPNotFound(HTTPException):
    """An error raised when a request is made to a non-existent endpoint."""

    def __init__(self, message: str) -> None:
        super().__init__(404, message)



"""Mafic CLI tools."""
# SPDX-License-Identifier: MIT

from __future__ import annotations

import platform
import sys
from argparse import ArgumentParser, Namespace

import aiohttp

import mafic

from .__libraries import library, version_info

VERSION_OUTPUT = """
- Python v{py.major}.{py.minor}.{py.micro}-{py.releaselevel}
- mafic v{mafic}
- aiohttp v{aiohttp}
- {lib} v{pkg.major}.{pkg.minor}.{pkg.micro}-{pkg.releaselevel}
- system info: {uname.system} {uname.release} {uname.version}
""".strip()


def show_version() -> None:
    """Show version information."""
    version = mafic.__version__
    uname = platform.uname()
    print(  # noqa: T201
        VERSION_OUTPUT.format(
            py=sys.version_info,
            mafic=version,
            aiohttp=aiohttp.__version__,
            uname=uname,
            lib=library,
            pkg=version_info,
        )
    )


def core(_: ArgumentParser, args: Namespace) -> None:
    """Core function for mafic CLI tools."""
    if args.version:
        show_version()


def parse_args() -> tuple[ArgumentParser, Namespace]:
    """Parse arguments from the CLI."""
    parser = ArgumentParser(prog="mafic", description="CLI tools for mafic")
    parser.add_argument(
        "-v",
        "--version",
        action="store_true",
        help="shows mafic and other related versions",
    )
    return parser, parser.parse_args()


def main() -> None:
    """Run the parser for arguments then execute."""
    parser, args = parse_args()
    core(parser, args)


if __name__ == "__main__":
    main()



"""A module contains region enums for voice regions and groups."""
# SPDX-License-Identifier: MIT

from __future__ import annotations

from enum import Enum

__all__ = ("Group", "Region", "VoiceRegion")


# This is taken from Vibr logs, javacord and nextcord enums and may not be complete.
# Discord only documents and exposes 13 of these regions, but there are more.
class VoiceRegion(Enum):
    """Represents a Discord voice region.

    All the attributes here are best effort but may not be complete.
    The value is what you find in ``region[number].discord.media`` endpoints.
    """

    OREGON = "oregon"
    MONTREAL = "montreal"
    SOUTH_AFRICA = "southafrica"
    JAPAN = "japan"
    DUBAI = "dubai"
    SANTA_CLARA = "santa-clara"
    FINLAND = "finland"
    MADRID = "madrid"
    ST_PETE = "st-pete"
    INDIA = "india"
    SEATTLE = "seattle"
    RUSSIA = "russia"
    SYDNEY = "sydney"
    HONG_KONG = "hongkong"
    MILAN = "milan"
    NEWARK = "newark"
    US_CENTRAL = "us-central"
    US_WEST = "us-west"
    SANTIAGO = "santiago"
    STOCKHOLM = "stockholm"
    BUENOS_AIRES = "buenos-aires"
    US_SOUTH = "us-south"
    US_EAST = "us-east"
    SINGAPORE = "singapore"
    BUCHAREST = "bucharest"
    ATLANTA = "atlanta"
    BRAZIL = "brazil"
    ROTTERDAM = "rotterdam"
    FRANKFURT = "frankfurt"
    LONDON = "london"
    SOUTH_KOREA = "south-korea"
    EUROPE = "europe"
    AMSTERDAM = "amsterdam"

    def __repr__(self) -> str:
        """Get the representation of this voice region."""
        return f"VoiceRegion.{self.name}"


class Region(Enum):
    """Represents a region of voice regions.

    This groups voice regions by geographical location.
    """

    EAST_NA = (VoiceRegion.MONTREAL, VoiceRegion.US_EAST, VoiceRegion.ATLANTA)
    CENTRAL_NA = (VoiceRegion.US_CENTRAL,)
    WEST_NA = (
        VoiceRegion.OREGON,
        VoiceRegion.SANTA_CLARA,
        VoiceRegion.SEATTLE,
        VoiceRegion.US_WEST,
    )
    SOUTH_NA = (VoiceRegion.US_SOUTH,)
    SOUTH_AMERICA = (VoiceRegion.SANTIAGO, VoiceRegion.BUENOS_AIRES, VoiceRegion.BRAZIL)
    SOUTH_AFRICA = (VoiceRegion.SOUTH_AFRICA,)
    NORTH_ASIA = (VoiceRegion.RUSSIA,)
    EAST_ASIA = (VoiceRegion.JAPAN, VoiceRegion.HONG_KONG)
    SOUTH_ASIA = (VoiceRegion.INDIA, VoiceRegion.SINGAPORE)
    WEST_ASIA = (VoiceRegion.DUBAI,)
    NORTH_EUROPE = (VoiceRegion.FINLAND, VoiceRegion.ST_PETE, VoiceRegion.STOCKHOLM)
    EAST_EUROPE = (VoiceRegion.BUCHAREST,)
    CENTRAL_EUROPE = (VoiceRegion.FRANKFURT, VoiceRegion.EUROPE)
    SOUTH_EUROPE = (VoiceRegion.MILAN,)
    WEST_EUROPE = (
        VoiceRegion.MADRID,
        VoiceRegion.NEWARK,
        VoiceRegion.ROTTERDAM,
        VoiceRegion.LONDON,
        VoiceRegion.AMSTERDAM,
    )
    OCEANIA = (VoiceRegion.SYDNEY,)

    def __repr__(self) -> str:
        """Get the representation of this region."""
        return f"<Region.{self.name}>"


class Group(Enum):
    """Represents a group of regions.

    This groups regions by geographical location. These are very generic and should
    be used rarely.
    """

    WEST = (
        Region.EAST_NA,
        Region.WEST_NA,
        Region.SOUTH_NA,
        Region.CENTRAL_NA,
        Region.SOUTH_AMERICA,
    )
    CENTRAL = (
        Region.NORTH_EUROPE,
        Region.WEST_EUROPE,
        Region.SOUTH_EUROPE,
        Region.EAST_EUROPE,
        Region.CENTRAL_EUROPE,
        Region.SOUTH_AFRICA,
    )
    EAST = (
        Region.NORTH_ASIA,
        Region.EAST_ASIA,
        Region.SOUTH_ASIA,
        Region.WEST_ASIA,
        Region.OCEANIA,
    )

    def __repr__(self) -> str:
        """Get the representation of this group."""
        return f"<Group.{self.name}>"



"""The module containing :class:`Playlist`."""
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .track import Track

if TYPE_CHECKING:
    from .typings import PlaylistInfo, TrackWithInfo

__all__ = ("Playlist",)


class Playlist:
    r"""Represents a playlist.

    Attributes
    ----------
    name: :class:`str`
        The name of the playlist.
    selected_track: :class:`int`
        The index of the selected track, if any.
    tracks: :class:`list`\[:class:`Track`]
        A list of tracks in the playlist.
    plugin_info: :class:`dict`\[:class:`str`, :class:`Any`]
        A dictionary containing plugin-specific information.

        .. versionadded:: 2.3
    """

    __slots__ = ("name", "plugin_info", "selected_track", "tracks")

    def __init__(
        self,
        *,
        info: PlaylistInfo,
        tracks: list[TrackWithInfo],
        plugin_info: dict[str, Any],
    ) -> None:
        self.name: str = info["name"]
        self.selected_track: int = info["selectedTrack"]
        self.tracks: list[Track] = [
            Track.from_data_with_info(track) for track in tracks
        ]
        self.plugin_info: dict[str, Any] = plugin_info



# SPDX-License-Identifier: MIT
# ruff: noqa: UP013
# `class` in `TypedDict` does not work as that is a reserved keyword

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, Optional, TypedDict, Union

if TYPE_CHECKING:
    from typing_extensions import NotRequired

    from .common import PlaylistInfo, TrackWithInfo
    from .misc import LavalinkException


__all__ = (
    "BalancingIPRouteDetails",
    "BalancingIPRoutePlanner",
    "BaseDetails",
    "ConfigureResumingResponse",
    "FailingIPAddress",
    "EmptyRoutePlanner",
    "Error",
    "GenericTracks",
    "Git",
    "IPBlock",
    "Info",
    "NanoIPRouteDetails",
    "NanoIPRoutePlanner",
    "NoMatches",
    "PlaylistTracks",
    "PluginData",
    "RotatingIPRouteDetails",
    "RotatingIPRoutePlanner",
    "RotatingNanoIPRouteDetails",
    "RotatingNanoIPRoutePlanner",
    "RoutePlannerStatus",
    "TrackLoadingResult",
    "TracksFailed",
    "Version",
    "Git",
    "Info",
    "RotatingIPRoutePlanner",
    "NanoIPRoutePlanner",
    "RotatingNanoIPRoutePlanner",
    "BalancingIPRoutePlanner",
    "EmptyRoutePlanner",
    "RoutePlannerStatus",
)


# V3


class PlaylistTracks(TypedDict):
    loadType: Literal["PLAYLIST_LOADED"]
    playlistInfo: PlaylistInfo
    tracks: list[TrackWithInfo]


class GenericTracks(TypedDict):
    loadType: Literal["TRACK_LOADED", "SEARCH_RESULT"]
    tracks: list[TrackWithInfo]


class TracksFailed(TypedDict):
    loadType: Literal["LOAD_FAILED"]
    exception: LavalinkException


class NoMatches(TypedDict):
    loadType: Literal["NO_MATCHES"]


# V4


class TrackLoadedV4(TypedDict):
    loadType: Literal["track"]
    data: TrackWithInfo


class PlaylistLoadedV4(TypedDict):
    loadType: Literal["playlist"]
    data: PlaylistDataV4


class PlaylistDataV4(TypedDict):
    info: PlaylistInfo
    pluginInfo: dict[str, Any]
    tracks: list[TrackWithInfo]


class SearchResultV4(TypedDict):
    loadType: Literal["search"]
    data: list[TrackWithInfo]


class SearchEmptyV4(TypedDict):
    loadType: Literal["empty"]


class LoadErrorV4(TypedDict):
    loadType: Literal["error"]
    data: LavalinkException


TrackLoadingResult = Union[
    PlaylistTracks,
    GenericTracks,
    TracksFailed,
    NoMatches,
    TrackLoadedV4,
    PlaylistLoadedV4,
    SearchResultV4,
    SearchEmptyV4,
    LoadErrorV4,
]


class PluginData(TypedDict):
    name: str
    version: str


class IPBlock(TypedDict):
    type: Literal["Inet4Address", "Inet6Address"]
    size: str


class FailingIPAddress(TypedDict):
    address: str
    failingTimestamp: int
    failingTime: str


class BaseDetails(TypedDict):
    ipBlock: IPBlock
    failingAddresses: list[FailingIPAddress]


class RotatingIPRouteDetails(BaseDetails):
    rotateIndex: str
    ipIndex: str
    currentAddress: str


# Fields are named class
RotatingIPRoutePlanner = TypedDict(
    "RotatingIPRoutePlanner",
    {
        "class": Literal["RotatingIpRoutePlanner"],
        "details": RotatingIPRouteDetails,
    },
)


class NanoIPRouteDetails(BaseDetails):
    currentAddressIndex: str


NanoIPRoutePlanner = TypedDict(
    "NanoIPRoutePlanner",
    {
        "class": Optional[Literal["NanoIpRoutePlanner"]],
        "details": Optional[NanoIPRouteDetails],
    },
)


class RotatingNanoIPRouteDetails(BaseDetails):
    blockIndex: str
    currentAddressIndex: str


RotatingNanoIPRoutePlanner = TypedDict(
    "RotatingNanoIPRoutePlanner",
    {
        "class": Optional[Literal["RotatingNanoIpRoutePlanner"]],
        "details": Optional[RotatingNanoIPRouteDetails],
    },
)


class BalancingIPRouteDetails(BaseDetails):
    ...


BalancingIPRoutePlanner = TypedDict(
    "BalancingIPRoutePlanner",
    {
        "class": Optional[Literal["BalancingIpRoutePlanner"]],
        "details": Optional[BalancingIPRouteDetails],
    },
)


EmptyRoutePlanner = TypedDict("EmptyRoutePlanner", {"class": None, "details": None})

RoutePlannerStatus = Union[
    RotatingIPRoutePlanner,
    NanoIPRoutePlanner,
    RotatingNanoIPRoutePlanner,
    BalancingIPRoutePlanner,
    EmptyRoutePlanner,
]


class ConfigureResumingResponse(TypedDict):
    resumingKey: str | None
    timeout: int


class Version(TypedDict):
    semver: str
    major: int
    minor: int
    patch: int
    preRelease: str | None


class Git(TypedDict):
    branch: str
    commit: str
    commitTime: int


class Info(TypedDict):
    version: Version
    buildTime: int
    git: Git
    jvm: str
    lavaplayer: str
    sourceManagers: list[str]
    filters: list[str]
    plugins: list[PluginData]


class Error(TypedDict):
    timestamp: int
    status: int
    error: str
    trace: NotRequired[str]
    message: str
    path: str



# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict, Union

from .common import Stats, TrackWithInfo
from .misc import PayloadWithGuild

if TYPE_CHECKING:
    from typing import Literal

    from typing_extensions import NotRequired

    from .misc import LavalinkException

__all__ = (
    "EventPayload",
    "PlayerUpdatePayload",
    "PlayerUpdateState",
    "IncomingMessage",
    "TrackEndEvent",
    "TrackExceptionEvent",
    "TrackStartEvent",
    "TrackStuckEvent",
    "WebSocketClosedEvent",
)


class PlayerUpdateState(TypedDict):
    time: int
    position: NotRequired[int]
    connected: bool
    ping: int


class PlayerUpdatePayload(PayloadWithGuild):
    op: Literal["playerUpdate"]
    state: PlayerUpdateState


class WebSocketClosedEvent(PayloadWithGuild):
    op: Literal["event"]
    type: Literal["WebSocketClosedEvent"]
    code: int
    reason: str
    byRemote: bool


class TrackStartEvent(PayloadWithGuild):
    op: Literal["event"]
    type: Literal["TrackStartEvent"]
    # V3
    encodedTrack: NotRequired[str]
    # V4
    track: NotRequired[TrackWithInfo]


class TrackEndEvent(PayloadWithGuild):
    op: Literal["event"]
    type: Literal["TrackEndEvent"]
    # V3
    encodedTrack: NotRequired[str]
    # V4
    track: NotRequired[TrackWithInfo]
    reason: Literal[
        # V3
        "FINISHED",
        "LOAD_FAILED",
        "STOPPED",
        "REPLACED",
        "CLEANUP",
        # V4
        "finished",
        "loadFailed",
        "stopped",
        "replaced",
        "cleanup",
    ]


class TrackExceptionEvent(PayloadWithGuild):
    op: Literal["event"]
    type: Literal["TrackExceptionEvent"]
    # V3
    encodedTrack: NotRequired[str]
    # V4
    track: NotRequired[TrackWithInfo]
    exception: LavalinkException


class TrackStuckEvent(PayloadWithGuild):
    op: Literal["event"]
    type: Literal["TrackStuckEvent"]
    # V3
    encodedTrack: NotRequired[str]
    # V4
    track: NotRequired[TrackWithInfo]
    thresholdMs: int


class ReadyPayload(TypedDict):
    op: Literal["ready"]
    resumed: bool
    sessionId: str


EventPayload = Union[
    WebSocketClosedEvent,
    TrackStartEvent,
    TrackEndEvent,
    TrackExceptionEvent,
    TrackStuckEvent,
]


IncomingMessage = Union[
    PlayerUpdatePayload,
    Stats,
    EventPayload,
    ReadyPayload,
]



# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Coroutine
from typing import Any, Literal, TypedDict, TypeVar

__all__ = (
    "Coro",
    "LavalinkException",
    "ExceptionSeverity",
    "PayloadWithGuild",
)
T = TypeVar("T")

Coro = Coroutine[Any, Any, T]
ExceptionSeverity = Literal[
    # V3
    "COMMON",
    "SUSPICIOUS",
    "FAULT",
    # V4
    "common",
    "suspicious",
    "fault",
]


class LavalinkException(TypedDict):
    severity: ExceptionSeverity
    message: str
    cause: str


class PayloadWithGuild(TypedDict):
    guildId: str



"""The strategy system for selecting a :class:`Node` from a :class:`NodePool`."""
# SPDX-License-Identifier: MIT

from __future__ import annotations

import re
from collections.abc import Callable
from enum import Enum, auto
from logging import getLogger
from random import choice
from typing import List, Optional

from .node import Node
from .region import VoiceRegion
from .type_variables import ClientT

StrategyCallable = Callable[
    [List[Node[ClientT]], int, Optional[int], Optional[str]], List[Node[ClientT]]
]
"""Represents a strategy callable.

This accepts ``nodes``, ``guild_id``, ``shard_count`` and ``endpoint`` and returns
a list of nodes.
"""

_log = getLogger(__name__)
__all__ = ("Strategy", "call_strategy")


class Strategy(Enum):
    """Represents a strategy for selecting a node."""

    LOCATION = auto()
    """Selects a node based on the region the guild is in."""

    RANDOM = auto()
    """Selects a random node."""

    SHARD = auto()
    """Selects a node based on the shard ID of the guild."""

    USAGE = auto()
    """Selects a node based on the least used node."""


def shard_strategy(
    nodes: list[Node[ClientT]], guild_id: int, shard_count: int | None, _: str | None
) -> list[Node[ClientT]]:
    """Get a node based on the shard ID of the guild.

    Parameters
    ----------
    nodes:
        The nodes to select from.
    guild_id:
        The ID of the guild to select a node for.
    shard_count:
        The number of shards the bot is using.
    _:
        Unused parameter.
    """
    if shard_count is None:
        shard_count = 1

    shard_id = (guild_id >> 22) % shard_count

    return list(
        filter(lambda node: node.shard_ids is None or shard_id in node.shard_ids, nodes)
    )


_REGION_REGEX = re.compile(r"(?:vip-)?(?P<region>[a-z-]{1,15})\d{1,5}\.discord\.media")


def location_strategy(
    nodes: list[Node[ClientT]], _: int, __: int | None, endpoint: str | None
) -> list[Node[ClientT]]:
    """Get a node based on the region the guild is in.

    Parameters
    ----------
    nodes:
        The nodes to select from.
    _:
        Unused parameter.
    __:
        Unused parameter.
    endpoint:
        The endpoint of the guild to select a node for.
    """
    if endpoint is None:
        return nodes

    match = _REGION_REGEX.match(endpoint)

    if not match:
        _log.error("Failed to find the region in an endpoint, defaulting to all nodes.")
        return nodes

    voice_region = match.group("region")

    if not voice_region:
        _log.error(
            "Failed to match endpoint %s (match: %s) to a region, "
            "defaulting to all nodes.",
            endpoint,
            match.group("region"),
        )
        return nodes

    try:
        voice_region = VoiceRegion(voice_region)
    except ValueError:
        _log.error(
            "Failed to match endpoint %s (match: %s) to a region, "
            "defaulting to all nodes.",
            endpoint,
            voice_region,
        )
        return nodes

    regional_nodes = list(
        filter(
            lambda node: node.regions is not None and voice_region in node.regions,
            nodes,
        )
    )

    if not regional_nodes:
        _log.error(
            "No nodes found for the region: %r, defaulting to all nodes.", voice_region
        )
        return nodes

    return regional_nodes


def usage_strategy(
    nodes: list[Node[ClientT]], _: int, __: int | None, ___: str | None
) -> list[Node[ClientT]]:
    """Get a node based on the least used node.

    This is calculated using :attr:`Node.weight`.

    Parameters
    ----------
    nodes:
        The nodes to select from.
    _:
        Unused parameter.
    __:
        Unused parameter.
    ___:
        Unused parameter.
    """
    # max() would be nice, however if all nodes have no stats, it returns the first.

    lowest = None

    for node in nodes:
        weight = node.weight

        if lowest is None or weight < lowest:
            lowest = weight

    if lowest is None:
        return nodes

    return list(filter(lambda node: node.weight == lowest, nodes))


def random_strategy(
    nodes: list[Node[ClientT]], _: int, __: int | None, ___: str | None
) -> list[Node[ClientT]]:
    """Get a random node.

    Parameters
    ----------
    nodes:
        The nodes to select from.
    _:
        Unused parameter.
    __:
        Unused parameter.
    ___:
        Unused parameter.
    """
    return [choice(nodes)]


# Used for ClientT preservation.
def call_strategy(
    strategy: Strategy,
    nodes: list[Node[ClientT]],
    guild_id: int,
    shard_count: int | None,
    endpoint: str | None,
) -> list[Node[ClientT]]:
    """Call a strategy's callback and get the result.

    Parameters
    ----------
    strategy:
        The strategy to call.
    nodes:
        The nodes to select from.
    guild_id:
        The ID of the guild to select a node for.
    shard_count:
        The number of shards the bot is using.
    endpoint:
        The endpoint of the voice client to select a node for.
    """
    if strategy is Strategy.LOCATION:
        return location_strategy(nodes, guild_id, shard_count, endpoint)
    elif strategy is Strategy.RANDOM:
        return random_strategy(nodes, guild_id, shard_count, endpoint)
    elif strategy is Strategy.SHARD:
        return shard_strategy(nodes, guild_id, shard_count, endpoint)
    elif strategy is Strategy.USAGE:
        return usage_strategy(nodes, guild_id, shard_count, endpoint)
    else:
        msg = f"Unknown strategy {strategy}"
        raise ValueError(msg)



r"""A module containing a :class:`NodePool`, used to manage :class:`Node`\s."""
# SPDX-License-Identifier: MIT

from __future__ import annotations

import asyncio
from collections.abc import Sequence
from functools import partial
from logging import getLogger
from random import choice
from typing import TYPE_CHECKING, Any, Generic, TypeVar, Union, cast

from .errors import NoNodesAvailable, PlayerNotConnected
from .node import Node
from .strategy import Strategy, StrategyCallable, call_strategy
from .type_variables import ClientT
from .utils import classproperty

if TYPE_CHECKING:
    from typing import ClassVar

    import aiohttp

    from .player import Player
    from .region import Group, Region, VoiceRegion


T = TypeVar("T")
__all__ = ("NodePool",)

_log = getLogger(__name__)


StrategyList = Union[
    StrategyCallable[ClientT],
    Sequence[Strategy],
    Sequence[StrategyCallable[ClientT]],
    Sequence[Union[Strategy, StrategyCallable[ClientT]]],
]
"""A type hint for a list of strategies to select the best node.

This can either be:

- A single :data:`.StrategyCallable`.
- A sequence of :class:`.Strategy`\\s.
- A sequence of :data:`.StrategyCallable`\\s.
- A sequence of :class:`.Strategy`\\s and :data:`.StrategyCallable`\\s.
"""


class NodePool(Generic[ClientT]):
    """A class that manages nodes and chooses them based on strategies.

    Parameters
    ----------
    client:
        The client to use to connect to the nodes.
    default_strategies:
        The default strategies to use when selecting a node. Defaults to
        [:attr:`Strategy.SHARD`, :attr:`Strategy.LOCATION`, :attr:`Strategy.USAGE`].
    """

    __slots__ = ()

    # Generics as expected, do not work in class variables.
    # Don't fear, the public methods using this are typed well.
    _nodes: ClassVar[dict[str, Node[Any]]] = {}
    _client: ClientT | None = None

    def __init__(
        self,
        client: ClientT,
        default_strategies: StrategyList[ClientT] | None = None,
    ) -> None:
        NodePool._client = client

        NodePool._default_strategies = (
            [
                Strategy.SHARD,
                Strategy.LOCATION,
                Strategy.USAGE,
            ]
            if default_strategies is None
            else default_strategies
        )

    @classproperty
    def label_to_node(cls) -> dict[str, Node[ClientT]]:
        """Get a mapping node labels to nodes."""
        return cls._nodes

    @classproperty
    def nodes(cls) -> list[Node[ClientT]]:
        """Get the list of all available nodes."""
        return list(filter(lambda n: n.available, cls._nodes.values()))

    async def create_node(
        self,
        *,
        host: str,
        port: int,
        label: str,
        password: str,
        secure: bool = False,
        heartbeat: int = 30,
        timeout: float = 10,
        session: aiohttp.ClientSession | None = None,
        resume_key: str | None = None,
        regions: Sequence[Group | Region | VoiceRegion] | None = None,
        shard_ids: Sequence[int] | None = None,
        resuming_session_id: str | None = None,
        player_cls: type[Player[ClientT]] | None = None,
    ) -> Node[ClientT]:
        r"""Create a node and connect it.

        The parameters here relate to :class:`Node`.

        Parameters
        ----------
        host:
            The host of the node to connect to.
        port:
            The port of the node to connect to.
        label:
            The label of the node used to identify it.
        password:
            The password of the node to authenticate.
        secure:
            Whether to use SSL (TLS) or not.
        heartbeat:
            The interval to send heartbeats to the node websocket connection.
        timeout:
            The timeout to use for the node websocket connection.
        session: :data:`~typing.Optional`\[:class:`aiohttp.ClientSession`]
            The session to use for the node websocket connection.
        resume_key:
            The key to use when resuming the node.
            If not provided, the key will be generated from the host, port and label.

            .. warning::

                This is ignored in lavalink V4, use ``resuming_session_id`` instead.
        regions:
            The voice regions that the node can be used in.
            This is used to determine when to use this node.
        shard_ids:
            The shard IDs that the node can be used in.
            This is used to determine when to use this node.
        resuming_session_id:
            The session ID to use when resuming the node.
            If not provided, the node will not resume.

            This should be stored from :func:`~mafic.on_node_ready` with
            :attr:`session_id` to resume the session and gain control of the players.
            If the node is not resuming, players will be destroyed if Lavalink loses
            connection to us.

            .. versionadded:: 2.2
        player_cls:
            The player class to use for this node when resuming.

            .. versionadded:: 2.8

        Returns
        -------
        :class:`Node`
            The created node.

        Raises
        ------
        RuntimeError
            If the node pool has not been initialized.
        """
        if self._client is None:
            msg = "NodePool has not been initialized."
            raise RuntimeError(msg)

        node = Node(
            host=host,
            port=port,
            label=label,
            password=password,
            client=self._client,
            secure=secure,
            heartbeat=heartbeat,
            timeout=timeout,
            session=session,
            resume_key=resume_key,
            regions=regions,
            shard_ids=shard_ids,
            resuming_session_id=resuming_session_id,
        )

        await self.add_node(node, player_cls=player_cls)
        return node

    async def add_node(
        self, node: Node[ClientT], *, player_cls: type[Player[ClientT]] | None = None
    ) -> None:
        """Add an existing node to this pool.

        .. note::

            You generally do not want this, use :meth:`create_node` instead.
            This is used for after running :meth:`remove_node` to re-add the node
            if it has been restarted.

        .. versionadded:: 2.7

        Parameters
        ----------
        node:
            The node to add.
        player:
            The player class to use for this node when resuming.

            .. versionadded:: 2.8
        """
        _log.info("Created node, connecting it...", extra={"label": node.label})
        await node.connect(player_cls=player_cls)

        self._nodes[node.label] = node

    async def remove_node(
        self, node: Node[ClientT] | str, *, transfer_players: bool = True
    ) -> None:
        """Remove a node from the pool.

        .. versionadded:: 2.6

        Parameters
        ----------
        node:
            The node to remove.
        transfer_players:
            Whether to transfer players to other nodes or destroy them.
        """
        if isinstance(node, str):
            node = self._nodes[node]

        # Remove prematurely so it is not chosen.
        del self._nodes[node.label]

        if transfer_players:

            async def transfer_player(player: Player[ClientT]) -> None:
                try:
                    target = self.get_node(
                        guild_id=player.guild.id,
                        endpoint=player.endpoint,  # pyright: ignore[reportPrivateUsage]
                    )
                    await player.transfer_to(target)
                except (RuntimeError, NoNodesAvailable, PlayerNotConnected):
                    _log.error(
                        "Failed to transfer player %d, destroying it...",
                        player.guild.id,
                        exc_info=True,
                        extra={"label": node.label},
                    )
                    await player.destroy()

            tasks = [transfer_player(player) for player in node.players]
            await asyncio.gather(*tasks)
        else:

            async def destroy_player(player: Player[ClientT]) -> None:
                _log.debug(
                    "Destroying player %d due to node removal...",
                    player.guild.id,
                    extra={"label": node.label},
                )
                await player.destroy()

            tasks = [destroy_player(player) for player in node.players]
            await asyncio.gather(*tasks)

        await node.close()

    async def close(self) -> None:
        """Close all nodes in the pool.

        .. versionadded:: 2.6
        """
        for node in self._nodes.values():
            await node.close()

    @classmethod
    def get_node(
        cls,
        *,
        guild_id: str | int,
        endpoint: str | None,
        strategies: StrategyList[ClientT] | None = None,
    ) -> Node[ClientT]:
        """Get a node based on the given strategies.

        Parameters
        ----------
        guild_id:
            The guild ID to get a node for.
        endpoint:
            The endpoint to get a node for.
        strategies:
            The strategies to use to get a node. If not provided, the default
            strategies will be used.

        Returns
        -------
        :class:`Node`
            The node to use.

        Raises
        ------
        RuntimeError
            If the node pool has not been initialized.
        """
        if cls._client is None:
            msg = "NodePool has not been initialized."
            raise RuntimeError(msg)

        actual_strategies: Sequence[StrategyCallable[ClientT] | Strategy]

        strategies = strategies or cls._default_strategies

        actual_strategies = [strategies] if callable(strategies) else strategies

        # It is a classproperty.
        # fmt: off
        nodes = cast("list[Node[ClientT]]", cls.nodes)  # pyright: ignore  # noqa: PGH003
        # fmt: on

        for strategy in actual_strategies:
            if isinstance(strategy, Strategy):
                strategy_callable = partial(call_strategy, strategy)
            else:
                strategy_callable = strategy

            nodes = strategy_callable(
                nodes, int(guild_id), cls._client.shard_count, endpoint
            )

            _log.debug(
                "Strategy %s returned nodes %s.",
                strategy.__name__ if callable(strategy) else strategy.name,
                ", ".join(n.label for n in nodes),
            )

            if len(nodes) == 1:
                return nodes[0]
            elif len(nodes) == 0:
                raise NoNodesAvailable

        return choice(nodes)

    @classmethod
    def get_random_node(cls) -> Node[ClientT]:
        """Get a random node.

        Returns
        -------
        :class:`Node`
            The random node.

        Raises
        ------
        ValueError
            If there are no nodes.
        """
        # It is a classproperty.
        nodes = cast(
            "list[Node[ClientT]]", cls.nodes  # pyright: ignore  # noqa: PGH003
        )

        if node := choice(nodes):
            return node

        raise NoNodesAvailable
